# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing the Neuroglancer viewer state."""

import collections
import collections.abc
import copy
import math
import numbers
import os
import re
import typing

import numpy as np
import numpy.typing

from . import local_volume, segment_colors, skeleton
from .coordinate_space import CoordinateArray, CoordinateSpace, DimensionScale
from .equivalence_map import EquivalenceMap
from .json_utils import encode_json_for_repr
from .json_wrappers import (
    JsonObjectWrapper,
    List,
    Map,
    _set_type_annotation,
    array_wrapper,
    bool_or_string,
    number_or_string,
    optional,
    typed_list,
    typed_map,
    typed_set,
    wrapped_property,
)

_BUILDING_DOCS = os.environ.get("NEUROGLANCER_BUILDING_DOCS") == "1"


__all__ = ["CoordinateSpace", "DimensionScale", "CoordinateArray"]


def export(obj):
    __all__.append(obj.__name__)
    return obj


T = typing.TypeVar("T")


def interpolate_linear(a, b, t):
    return a * (1 - t) + b * t


def interpolate_linear_optional_vectors(a, b, t):
    if a is not None and b is not None and len(a) == len(b):
        return a * (1 - t) + b * t
    return a


def unit_quaternion():
    return np.array([0, 0, 0, 1], np.float32)


def quaternion_slerp(a, b, t):
    """Spherical linear interpolation for unit quaternions.

    This is based on the implementation in the gl-matrix package:
    https://github.com/toji/gl-matrix
    """
    if a is None:
        a = unit_quaternion()
    if b is None:
        b = unit_quaternion()
    # calc cosine
    cosom = np.dot(a, b)
    # adjust signs (if necessary)
    if cosom < 0.0:
        cosom = -cosom
        b = -b

    # calculate coefficients
    if (1.0 - cosom) > 0.000001:
        # standard case (slerp)
        omega = math.acos(cosom)
        sinom = math.sin(omega)
        scale0 = math.sin((1.0 - t) * omega) / sinom
        scale1 = math.sin(t * omega) / sinom
    else:
        # "from" and "to" quaternions are very close
        #  ... so we can do a linear interpolation
        scale0 = 1.0 - t
        scale1 = t
    return scale0 * a + scale1 * b


def interpolate_zoom(a, b, t):
    if a is None or b is None:
        return a
    scale_change = math.log(b / a)
    return a * math.exp(scale_change * t)


class _ToolMetaclass(type):
    def __call__(self, obj: typing.Any = None, _readonly: bool = False, **kwargs):
        return _factory_new(
            registry=tool_types,
            base_class=Tool,
            cls=self,
            allow_str=True,
            obj=obj,
            _readonly=_readonly,
            kwargs=kwargs,
        )


@export
class Tool(JsonObjectWrapper, metaclass=_ToolMetaclass):
    __slots__ = ()

    type = wrapped_property("type", str)

    TOOL_TYPE: str

    def __init__(self, *args, **kwargs):
        tool_type = self.TOOL_TYPE
        if tool_type is not None:
            kwargs.update(type=tool_type)
        super().__init__(*args, **kwargs)

    def __new__(cls, json_data=None, _readonly: bool = False, **kwargs):
        """Coerces the argument to a `Tool`."""
        return object.__new__(cls)


tool_types = {}


def export_tool(tool_class):
    export(tool_class)
    tool_types[tool_class.TOOL_TYPE] = tool_class
    return tool_class


@export
class LayerTool(Tool):
    __slots__ = ()

    layer = wrapped_property("layer", optional(str))
    """Name of the layer to which this tool applies.

    Only valid for tools contained within `~ToolPalette.tools`.
    """


@export_tool
class PlacePointTool(Tool):
    __slots__ = ()
    TOOL_TYPE = "annotatePoint"


@export_tool
class PlaceLineTool(Tool):
    __slots__ = ()
    TOOL_TYPE = "annotateLine"


@export_tool
class PlaceBoundingBoxTool(Tool):
    __slots__ = ()
    TOOL_TYPE = "annotateBoundingBox"


@export_tool
class PlaceEllipsoidTool(Tool):
    __slots__ = ()
    TOOL_TYPE = "annotateSphere"


@export_tool
class BlendTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "blend"


@export_tool
class OpacityTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "opacity"


@export_tool
class VolumeRenderingTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "volumeRendering"


@export_tool
class VolumeRenderingGainTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "volumeRenderingGain"


@export_tool
class VolumeRenderingDepthSamplesTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "volumeRenderingDepthSamples"


@export_tool
class CrossSectionRenderScaleTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "crossSectionRenderScale"


@export_tool
class SelectedAlphaTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "selectedAlpha"


@export_tool
class NotSelectedAlphaTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "notSelectedAlpha"


@export_tool
class ObjectAlphaTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "objectAlpha"


@export_tool
class HideSegmentZeroTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "hideSegmentZero"


@export_tool
class HoverHighlightTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "hoverHighlight"


@export_tool
class BaseSegmentColoringTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "baseSegmentColoring"


@export_tool
class IgnoreNullVisibleSetTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "ignoreNullVisibleSet"


@export_tool
class ColorSeedTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "colorSeed"


@export_tool
class SegmentDefaultColorTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "segmentDefaultColor"


@export_tool
class MeshRenderScaleTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "meshRenderScale"


@export_tool
class MeshSilhouetteRenderingTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "meshSilhouetteRendering"


@export_tool
class SaturationTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "saturation"


@export_tool
class SkeletonRenderingMode2dTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "skeletonRendering.mode2d"


@export_tool
class SkeletonRenderingMode3dTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "skeletonRendering.mode3d"


@export_tool
class SkeletonRenderingLineWidth2dTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "skeletonRendering.lineWidth2d"


@export_tool
class SkeletonRenderingLineWidth3dTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "skeletonRendering.lineWidth3d"


@export_tool
class ShaderControlTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "shaderControl"
    control = wrapped_property("control", str)


@export_tool
class MergeSegmentsTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "mergeSegments"


@export_tool
class SplitSegmentsTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "splitSegments"


@export_tool
class SelectSegmentsTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "selectSegments"


@export_tool
class DimensionTool(LayerTool):
    __slots__ = ()
    TOOL_TYPE = "dimension"
    dimension = wrapped_property("dimension", str)


@export
class SidePanelLocation(JsonObjectWrapper):
    __slots__ = ()
    side = wrapped_property("side", optional(str))
    visible = wrapped_property("visible", optional(bool))
    size = wrapped_property("size", optional(int))
    flex = wrapped_property("flex", optional(float, 1))
    row = wrapped_property("row", optional(int))
    col = wrapped_property("col", optional(int))


@export
class ToolPalette(SidePanelLocation):
    __slots__ = ()
    tools = wrapped_property("tools", typed_list(Tool))
    query = wrapped_property("query", optional(str))


@export
class SelectedLayerState(SidePanelLocation):
    __slots__ = ()
    layer = wrapped_property("layer", optional(str))


@export
class StatisticsDisplayState(SidePanelLocation):
    pass


@export
class LayerSidePanelState(SidePanelLocation):
    tab = wrapped_property("tab", optional(str))
    tabs = wrapped_property("tabs", typed_set(str))


@export
class LayerListPanelState(SidePanelLocation):
    pass


@export
class HelpPanelState(SidePanelLocation):
    pass


@export
class DimensionPlaybackVelocity(JsonObjectWrapper):
    __slot__ = ()
    supports_validation = True

    velocity = wrapped_property("velocity", optional(float, 10))
    at_boundary = atBoundary = wrapped_property("atBoundary", optional(str, "reverse"))
    paused = wrapped_property("paused", optional(bool, True))


@export
class Layer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", optional(str))
    local_dimensions = localDimensions = wrapped_property(
        "localDimensions", CoordinateSpace
    )
    local_position = localPosition = wrapped_property(
        "localPosition", optional(array_wrapper(np.float32))
    )
    local_velocity = localVelocity = wrapped_property(
        "localVelocity", typed_map(key_type=str, value_type=DimensionPlaybackVelocity)
    )

    tab = wrapped_property("tab", optional(str))
    panels = wrapped_property("panels", typed_list(LayerSidePanelState))
    pick = wrapped_property("pick", optional(bool))
    tool_bindings = wrapped_property(
        "toolBindings", typed_map(key_type=str, value_type=Tool)
    )
    tool = wrapped_property("tool", optional(Tool))

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.layer_position = interpolate_linear_optional_vectors(
            a.layer_position, b.layer_position, t
        )
        return c


@export
class PointAnnotationLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="pointAnnotation", **kwargs)

    points = wrapped_property("points", typed_list(array_wrapper(np.float32, 3)))


@export
class CoordinateSpaceTransform(JsonObjectWrapper):
    __slots__ = ()

    output_dimensions = outputDimensions = wrapped_property(
        "outputDimensions", CoordinateSpace
    )
    input_dimensions = inputDimensions = wrapped_property(
        "inputDimensions", optional(CoordinateSpace)
    )
    source_rank = sourceRank = wrapped_property("sourceRank", optional(int))
    matrix = wrapped_property("matrix", optional(array_wrapper(np.float64)))


def data_source_url(x):
    if isinstance(x, local_volume.LocalVolume | skeleton.SkeletonSource):
        return x
    if not isinstance(x, str):
        raise TypeError
    return x


@export
class LayerDataSubsource(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True

    def __init__(self, json_data=None, *args, **kwargs):
        if isinstance(json_data, bool):
            json_data = {"enabled": json_data}
        super().__init__(json_data, *args, **kwargs)

    enabled = wrapped_property("enabled", optional(bool))


@export
class LayerDataSource(JsonObjectWrapper):
    __slots__ = ()

    def __init__(self, json_data=None, *args, **kwargs):
        if isinstance(json_data, str) or isinstance(
            json_data, local_volume.LocalVolume | skeleton.SkeletonSource
        ):
            json_data = {"url": json_data}
        super().__init__(json_data, *args, **kwargs)

    url = wrapped_property("url", data_source_url)
    transform = wrapped_property("transform", optional(CoordinateSpaceTransform))
    subsources = wrapped_property("subsources", typed_map(str, LayerDataSubsource))
    enable_default_subsources = enableDefaultSubsources = wrapped_property(
        "enableDefaultSubsources", optional(bool, True)
    )


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LayerDataSourcesBase = List[LayerDataSource]
else:
    _LayerDataSourcesBase = typed_list(LayerDataSource, validator=LayerDataSource)


@export
class LayerDataSources(_LayerDataSourcesBase):
    __slots__ = ()

    def __init__(self, json_data=None, **kwargs):
        if isinstance(
            json_data,
            LayerDataSource
            | str
            | local_volume.LocalVolume
            | skeleton.SkeletonSource
            | dict,
        ):
            json_data = [json_data]
        elif isinstance(json_data, LayerDataSources):
            json_data = json_data.to_json()
        super().__init__(json_data, **kwargs)


class _AnnotationLayerOptions:
    __slots__ = ()
    annotation_color = annotationColor = wrapped_property(
        "annotationColor", optional(str)
    )


@export
class InvlerpParameters(JsonObjectWrapper):
    range = wrapped_property("range", optional(array_wrapper(numbers.Number, 2)))
    window = wrapped_property("window", optional(array_wrapper(numbers.Number, 2)))
    channel = wrapped_property("channel", optional(typed_list(int)))


@export
class TransferFunctionParameters(JsonObjectWrapper):
    window = wrapped_property("window", optional(array_wrapper(numbers.Number, 2)))
    channel = wrapped_property("channel", optional(typed_list(int)))
    controlPoints = wrapped_property(
        "controlPoints", optional(typed_list(typed_list(number_or_string)))
    )
    defaultColor = wrapped_property("defaultColor", optional(str))


_UINT64_STR_PATTERN = re.compile("[0-9]+")


def _shader_control_parameters(v, _readonly=False):
    if isinstance(v, str):
        # Check if it can be converted to a number
        if _UINT64_STR_PATTERN.fullmatch(v):
            return int(v)
        return v
    if isinstance(v, numbers.Number):
        return v
    if isinstance(v, dict):
        if "controlPoints" in v:
            return TransferFunctionParameters(v, _readonly=_readonly)
        return InvlerpParameters(v, _readonly=_readonly)
    if isinstance(v, InvlerpParameters):
        return v
    if isinstance(v, TransferFunctionParameters):
        return v
    raise TypeError(f"Unexpected shader control parameters type: {type(v)}")


_set_type_annotation(
    _shader_control_parameters,
    numbers.Number | str | InvlerpParameters | TransferFunctionParameters,
)


_shader_control_parameters.supports_readonly = True  # type: ignore[attr-defined]

ShaderControls = typed_map(str, _shader_control_parameters)


@export
class ImageLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="image", **kwargs)

    source = wrapped_property("source", LayerDataSources)
    shader = wrapped_property("shader", str)
    shader_controls = shaderControls = wrapped_property(
        "shaderControls", ShaderControls
    )
    opacity = wrapped_property("opacity", optional(float, 0.5))
    blend = wrapped_property("blend", optional(str))
    volume_rendering_mode = volumeRenderingMode = VolumeRendering = volume_rendering = (
        wrapped_property("volumeRendering", optional(bool_or_string, False))
    )
    volume_rendering_gain = volumeRenderingGain = wrapped_property(
        "volumeRenderingGain", optional(float, 0)
    )
    volume_rendering_depth_samples = volumeRenderingDepthSamples = wrapped_property(
        "volumeRenderingDepthSamples", optional(float, 64)
    )
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        "crossSectionRenderScale", optional(float, 1)
    )

    @staticmethod
    def interpolate(a, b, t):
        c = Layer.interpolate(a, b, t)
        c.opacity = interpolate_linear(a.opacity, b.opacity, t)
        return c


def uint64_equivalence_map(obj, _readonly=False):
    if isinstance(obj, EquivalenceMap):
        return obj
    if obj is not None:
        obj = [[int(v) for v in group] for group in obj]
    return EquivalenceMap(obj, _readonly=_readonly)


def _linked_segmentation_color_group_value(x):
    if isinstance(x, str):
        return x
    if x is False:
        return x
    raise ValueError(f"Expected str or False, but received: {x!r}")


@export
class SkeletonRenderingOptions(JsonObjectWrapper):
    __slots__ = ()

    shader = wrapped_property("shader", optional(str))
    shader_controls = shaderControls = wrapped_property(
        "shaderControls", ShaderControls
    )
    mode2d = wrapped_property("mode2d", optional(str))
    line_width2d = lineWidth2d = wrapped_property("lineWidth2d", optional(float, 2))
    mode3d = wrapped_property("mode3d", optional(str))
    line_width3d = lineWidth3d = wrapped_property("lineWidth3d", optional(float, 1))


@export
class StarredSegments(collections.abc.MutableMapping[int, bool]):
    supports_readonly = True
    supports_validation = True
    __slots__ = ("_readonly", "_data", "_visible")

    _readonly: bool
    _data: dict[int, bool]
    _visible: dict[int, bool]

    def __init__(self, json_data=None, _readonly=False):
        self._readonly = _readonly
        self._data = {}
        self._visible = {}
        if json_data is None:
            return
        self._update(json_data)

    def _update(self, other):
        if isinstance(other, StarredSegments):
            self._data.update(other._data)
            visible = self._visible
            for k, v in other._data:
                if not v:
                    visible.pop(k, None)
            visible.update(other._visible)
            return

        if isinstance(other, collections.abc.Mapping):
            items = other.items()
        else:
            items = other
        data = self._data
        visible = self._visible
        for item in items:
            if isinstance(item, numbers.Integral):
                k = np.uint64(item)
                if k != item:
                    raise ValueError(f"Invalid uint64: {item!r}")
                v = True
            elif isinstance(item, str):
                v = True
                if item.startswith("!"):
                    v = False
                    item = item[1:]
                k = np.uint64(item)
            elif isinstance(item, tuple):
                k, v = item
                if (
                    not isinstance(k, numbers.Integral)
                    or np.uint64(k) != k
                    or not isinstance(v, bool)
                ):
                    raise TypeError(f"Invalid (uint64, bool) pair: {(k, v)!r}")
                k = np.uint64(k)
            else:
                raise TypeError(
                    f"Expected int | str | Tuple[uint64, bool] but received: {item!r}"
                )

            data[k] = v
            if v:
                visible[k] = True
            else:
                visible.pop(k, None)

    def copy(self):
        """Returns a copy of the starred segment list."""
        return StarredSegments(self)

    def __len__(self) -> int:
        """Returns the number of starred segments."""
        return len(self._data)

    def __contains__(self, segment_id: typing.Any) -> bool:
        """Checks if a segment is starred."""
        return segment_id in self._data

    def keys(self) -> collections.abc.KeysView[int]:
        """Returns a view of the starred segments."""
        return self._data.keys()

    def values(self) -> collections.abc.ValuesView[bool]:
        """Returns a view of the visibility state of each starred segment."""
        return self._data.values()

    def items(self) -> collections.abc.ItemsView[int, bool]:
        """Returns a view of the (segment, visible) pairs."""
        return self._data.items()

    def __eq__(self, other) -> bool:
        if isinstance(other, StarredSegments):
            return self._data == other._data
        return self._data == other

    def add(self, segment_id: int) -> None:
        """Adds a starred segment, marking it visible if not already starred."""
        if self._readonly:
            raise AttributeError
        self.setdefault(segment_id, True)

    @typing.overload
    def get(self, segment_id: int) -> bool | None: ...

    @typing.overload
    def get(self, segment_id: int, default: T) -> bool | T: ...

    def get(self, segment_id: int, default=None):
        """Checks if a segment is visible.

        Args:
          segment_id: Segment to check.
          default: Return value if :py:param:`.segment_id` is not starred.

        Returns:
          `True` if visible, `False` if starred but not visible,
          :py:param:`.default` if not starred.
        """
        return self._data.get(segment_id, default)

    def __getitem__(self, segment_id: int) -> bool:
        """Checks if a starred segment is visible."""
        return self._data[segment_id]

    def remove(self, segment_id: int) -> None:
        """Removes a segment from the starred list.

        Raises:
          KeyError: if the segment is not starred.
        """
        if self._readonly:
            raise AttributeError
        del self._data[segment_id]
        self._visible.pop(segment_id)

    def discard(self, segment_id: int) -> None:
        """Removes a segment from the starred list if present."""
        self._data.pop(segment_id, None)
        self._visible.pop(segment_id, None)

    def __setitem__(self, segment_id: int, visible: bool) -> None:
        """Stars and sets the visibility of a segment."""
        if self._readonly:
            raise AttributeError
        self._data[segment_id] = visible
        if visible:
            self._visible[segment_id] = True
        else:
            self._visible.pop(segment_id, None)

    def __delitem__(self, segment_id: int) -> None:
        """Removes a segment from the starred list.

        Raises:
          KeyError: if the segment is not starred.
        """
        if self._readonly:
            raise AttributeError
        del self._data[segment_id]
        self._visible.pop(segment_id, None)

    def clear(self):
        """Unstars all segments."""
        if self._readonly:
            raise AttributeError
        self._data.clear()
        self._visible.clear()

    def __repr__(self):
        return f"StarredSegments({self._data!r})"

    def update(  # type: ignore[override]
        self,
        other: typing.Union[
            "StarredSegments",
            collections.abc.MutableMapping[int, bool],
            typing.Iterable[int | str | tuple[int, bool]],
        ],
    ):
        """Merges in additional starred segments."""
        if self._readonly:
            raise AttributeError
        self._update(other)

    def to_json(self) -> list[str]:
        """Returns the representation as a list of strings."""
        return [
            f"{segment}" if visible else f"!{segment}"
            for segment, visible in self.items()
        ]

    def __iter__(self) -> typing.Iterator[int]:
        return iter(self._data)

    @property
    def visible(self) -> "VisibleSegments":
        return VisibleSegments(self)

    @visible.setter
    def visible(self, segments: collections.abc.Iterable[int]):
        new_dict = {}
        for k in segments:
            num_k = np.uint64(k)
            if num_k != k:
                raise ValueError(f"Invalid uint64 value: {k}")
            new_dict[int(num_k)] = True
        self._data = new_dict
        self._visible = new_dict.copy()


@export
class VisibleSegments(collections.abc.MutableSet[int]):
    """Subset of visible segments within a `StarredSegments` object."""

    def __init__(self, starred_segments: StarredSegments):
        """Constructs a view of the visible segments within a ``StarredSegments`` object."""
        self._starred_segments = starred_segments
        self._visible = self._starred_segments._visible

    def __len__(self) -> int:
        """Returns the number of visible segments."""
        return len(self._visible)

    def clear(self):
        """Unstars all segments."""
        self._starred_segments.clear()

    def __contains__(self, segment_id: typing.Any) -> bool:
        """Checks if a segment is visible."""
        return segment_id in self._visible

    def add(self, segment_id: int) -> None:
        """Stars a segment and marks it visible."""
        self._starred_segments[segment_id] = True

    def discard(self, segment_id) -> None:
        """Unstars a segment if present."""
        self._starred_segments.discard(segment_id)

    def __iter__(self) -> typing.Iterator[int]:
        """Iterates over the visible segments."""
        return iter(self._visible)

    def copy(self) -> "VisibleSegments":
        """Returns a copy of the visible segment list."""
        new_starred_segments = StarredSegments()
        new_visible = self._visible.copy()
        new_starred_segments._data = new_visible
        new_starred_segments._visible = new_visible.copy()
        return VisibleSegments(new_starred_segments)

    def __repr__(self):
        return f"VisibleSegments({list(self)!r})"


@export
class SegmentationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="segmentation", **kwargs)

    source = wrapped_property("source", LayerDataSources)
    starred_segments = wrapped_property("segments", StarredSegments)

    @property
    def visible_segments(self):
        return VisibleSegments(self.starred_segments)

    @visible_segments.setter
    def visible_segments(self, segments):
        self.starred_segments.visible = segments

    segments = visible_segments

    equivalences = wrapped_property("equivalences", uint64_equivalence_map)
    hide_segment_zero = hideSegmentZero = wrapped_property(
        "hideSegmentZero", optional(bool, True)
    )
    hover_highlight = hoverHighlight = wrapped_property(
        "hoverHighlight", optional(bool, True)
    )
    base_segment_coloring = baseSegmentColoring = wrapped_property(
        "baseSegmentColoring", optional(bool, False)
    )
    selected_alpha = selectedAlpha = wrapped_property(
        "selectedAlpha", optional(float, 0.5)
    )
    not_selected_alpha = notSelectedAlpha = wrapped_property(
        "notSelectedAlpha", optional(float, 0)
    )
    object_alpha = objectAlpha = wrapped_property("objectAlpha", optional(float, 1.0))
    saturation = wrapped_property("saturation", optional(float, 1.0))
    ignore_null_visible_set = ignoreNullVisibleSet = wrapped_property(
        "ignoreNullVisibleSet", optional(bool, True)
    )
    skeleton_rendering = skeletonRendering = wrapped_property(
        "skeletonRendering", SkeletonRenderingOptions
    )

    @property
    def skeleton_shader(self):
        return self.skeleton_rendering.shader

    @skeleton_shader.setter
    def skeleton_shader(self, shader):
        self.skeleton_rendering.shader = shader

    skeletonShader = skeleton_shader

    color_seed = colorSeed = wrapped_property("colorSeed", optional(int, 0))
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        "crossSectionRenderScale", optional(float, 1)
    )
    mesh_render_scale = meshRenderScale = wrapped_property(
        "meshRenderScale", optional(float, 10)
    )
    mesh_silhouette_rendering = meshSilhouetteRendering = wrapped_property(
        "meshSilhouetteRendering", optional(float, 0)
    )
    segment_query = segmentQuery = wrapped_property("segmentQuery", optional(str))
    segment_colors = segmentColors = wrapped_property(
        "segmentColors", typed_map(key_type=np.uint64, value_type=str)
    )
    segment_default_color = segmentDefaultColor = wrapped_property(
        "segmentDefaultColor", optional(str)
    )

    @property
    def segment_html_color_dict(self):
        """Returns a dictionary whose keys are segments and values are the 6-digit hex
        strings representing the colors of those segments given the current
        color seed
        """
        d = {}
        for segment in self.segments:
            hex_string = segment_colors.hex_string_from_segment_id(
                color_seed=self.color_seed, segment_id=segment
            )
            d[segment] = hex_string
        return d

    linked_segmentation_group = linkedSegmentationGroup = wrapped_property(
        "linkedSegmentationGroup", optional(str)
    )
    linked_segmentation_color_group = linkedSegmentationColorGroup = wrapped_property(
        "linkedSegmentationColorGroup", optional(_linked_segmentation_color_group_value)
    )

    @staticmethod
    def interpolate(a, b, t):
        c = Layer.interpolate(a, b, t)
        for k in ["selected_alpha", "not_selected_alpha", "object_alpha"]:
            setattr(c, k, interpolate_linear(getattr(a, k), getattr(b, k), t))
        return c


@export
class SingleMeshLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="mesh", **kwargs)

    source = wrapped_property("source", LayerDataSources)
    vertex_attribute_sources = vertexAttributeSources = wrapped_property(
        "vertexAttributeSources", optional(typed_list(str))
    )
    shader = wrapped_property("shader", str)
    vertex_attribute_names = vertexAttributeNames = wrapped_property(
        "vertexAttributeNames", optional(typed_list(optional(str)))
    )


def _factory_new(
    registry: collections.abc.Mapping[str, type],
    base_class: type,
    cls: type,
    allow_str: bool,
    obj: typing.Any,
    _readonly: bool,
    kwargs,
):
    if cls is base_class:
        if isinstance(obj, base_class):
            cls = type(obj)
        else:
            if allow_str and isinstance(obj, str):
                t = obj
                obj = {"type": obj}
            else:
                if not isinstance(obj, dict):
                    raise TypeError("Expected dict", obj)
                t = obj.get("type")  # type: ignore[assignment]
            cls = registry[t]  # type: ignore[index]
    return type.__call__(cls, obj, _readonly=_readonly, **kwargs)


class _AnnotationMetaclass(type):
    def __call__(self, obj: typing.Any = None, _readonly: bool = False, **kwargs):
        return _factory_new(
            registry=annotation_types,
            base_class=Annotation,
            cls=self,
            allow_str=False,
            obj=obj,
            _readonly=_readonly,
            kwargs=kwargs,
        )


class Annotation(JsonObjectWrapper, metaclass=_AnnotationMetaclass):
    __slots__ = ()

    id = wrapped_property("id", optional(str))  # pylint: disable=invalid-name
    type = wrapped_property("type", str)
    description = wrapped_property("description", optional(str))
    segments = wrapped_property("segments", optional(typed_list(typed_list(np.uint64))))
    props = wrapped_property("props", optional(typed_list(number_or_string)))

    def __new__(cls, obj=None, _readonly: bool = False, **kwargs):
        """Coerces the argument to an `Annotation`."""
        return object.__new__(cls)


@export
class PointAnnotation(Annotation):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="point", **kwargs)

    point = wrapped_property("point", array_wrapper(np.float32))


@export
class LineAnnotation(Annotation):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="line", **kwargs)

    point_a = pointA = wrapped_property("pointA", array_wrapper(np.float32))
    point_b = pointB = wrapped_property("pointB", array_wrapper(np.float32))


@export
class AxisAlignedBoundingBoxAnnotation(Annotation):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="axis_aligned_bounding_box", **kwargs)

    point_a = pointA = wrapped_property("pointA", array_wrapper(np.float32))
    point_b = pointB = wrapped_property("pointB", array_wrapper(np.float32))


@export
class EllipsoidAnnotation(Annotation):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="ellipsoid", **kwargs)

    center = wrapped_property("center", array_wrapper(np.float32))
    radii = wrapped_property("radii", array_wrapper(np.float32))


annotation_types = {
    "point": PointAnnotation,
    "line": LineAnnotation,
    "axis_aligned_bounding_box": AxisAlignedBoundingBoxAnnotation,
    "ellipsoid": EllipsoidAnnotation,
}


@export
class AnnotationPropertySpec(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property("id", str)
    type = wrapped_property("type", str)
    description = wrapped_property("description", optional(str))
    default = wrapped_property("default", optional(number_or_string))
    enum_values = wrapped_property(
        "enum_values", optional(typed_list(number_or_string))
    )
    enum_labels = wrapped_property("enum_labels", optional(typed_list(str)))


@export
class AnnotationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, type="annotation", **kwargs)

    source = wrapped_property("source", LayerDataSources)
    annotations = wrapped_property("annotations", typed_list(Annotation))
    annotation_properties = annotationProperties = wrapped_property(
        "annotationProperties", typed_list(AnnotationPropertySpec)
    )
    annotation_relationships = annotationRelationships = wrapped_property(
        "annotationRelationships", typed_list(str)
    )
    linked_segmentation_layer = linkedSegmentationLayer = wrapped_property(
        "linkedSegmentationLayer", typed_map(str, str)
    )
    filter_by_segmentation = filterBySegmentation = wrapped_property(
        "filterBySegmentation", typed_list(str)
    )
    ignore_null_segment_filter = ignoreNullSegmentFilter = wrapped_property(
        "ignoreNullSegmentFilter", optional(bool, True)
    )
    shader = wrapped_property("shader", str)
    shader_controls = shaderControls = wrapped_property(
        "shaderControls", ShaderControls
    )

    @staticmethod
    def interpolate(a, b, t):
        del b
        del t
        return a


@export
class LocalAnnotationLayer(AnnotationLayer):
    __slots__ = ()

    def __init__(self, dimensions, *args, **kwargs):
        super().__init__(
            *args,
            source=LayerDataSource(
                url="local://annotations",
                transform=CoordinateSpaceTransform(outputDimensions=dimensions),
            ),
            **kwargs,
        )


layer_types = {
    "image": ImageLayer,
    "segmentation": SegmentationLayer,
    "pointAnnotation": PointAnnotationLayer,
    "annotation": AnnotationLayer,
    "mesh": SingleMeshLayer,
}


def make_layer(json_data, _readonly=False):
    if isinstance(json_data, Layer):
        return json_data

    if isinstance(json_data, local_volume.LocalVolume):
        json_data = dict(type=json_data.volume_type, source=json_data)

    if not isinstance(json_data, dict):
        raise TypeError

    type_name = json_data.get("type")
    layer_type = layer_types.get(type_name)
    if layer_type is not None:
        return layer_type(json_data, _readonly=_readonly)
    else:
        raise ValueError


@export
class ManagedLayer(JsonObjectWrapper):
    __slots__ = ("name", "layer")

    def __init__(self, name, layer=None, _readonly=False, **kwargs):
        if isinstance(name, ManagedLayer):
            if layer is not None or kwargs:
                raise ValueError
            layer = name.to_json()
            name = name.name

        object.__setattr__(self, "name", name)

        if isinstance(layer, Layer):
            json_data = {}
        elif isinstance(layer, local_volume.LocalVolume):
            json_data = {}
            layer = make_layer(layer, _readonly=_readonly)
        else:
            if layer is None:
                json_data = {}
            else:
                json_data = layer
            layer = make_layer(json_data, _readonly=_readonly)

        object.__setattr__(self, "layer", layer)
        super().__init__(json_data, _readonly=_readonly, **kwargs)

    _visible = wrapped_property("visible", optional(bool))
    archived = wrapped_property("archived", optional(bool, False))

    @property
    def visible(self):
        return not self.archived and self._visible is not False

    @visible.setter
    def visible(self, value):
        self._visible = value

    def __getattr__(self, key):
        return getattr(self.layer, key)

    def __setattr__(self, key, value):
        if self._readonly:
            raise AttributeError
        if key in ["name", "_visible", "visible", "archived", "layer"]:
            object.__setattr__(self, key, value)
        else:
            return setattr(self.layer, key, value)

    def __repr__(self):
        return f"ManagedLayer({encode_json_for_repr(self.name)},{encode_json_for_repr(self.to_json())})"

    def to_json(self):
        r = self.layer.to_json()
        r["name"] = self.name
        archived = self.archived
        if not archived:
            r.pop("archived", None)
        else:
            r["archived"] = True
        visible = self.visible
        if visible or archived:
            r.pop("visible", None)
        else:
            r["visible"] = False
        return r

    def __deepcopy__(self, memo):
        return ManagedLayer(self.name, copy.deepcopy(self.to_json(), memo))


@export
class Layers:
    __slots__ = ("_layers", "_readonly")
    supports_readonly = True

    def __init__(self, json_data, _readonly=False):
        if json_data is None:
            json_data = {}
        self._layers = []
        self._readonly = _readonly
        if isinstance(json_data, collections.abc.Mapping):
            for k, v in json_data.items():
                self._layers.append(ManagedLayer(k, v, _readonly=_readonly))
        else:
            # layers property can also be an array in JSON now. each layer has a name property
            for layer in json_data:
                if isinstance(layer, ManagedLayer):
                    self._layers.append(
                        ManagedLayer(layer.name, layer, _readonly=_readonly)
                    )
                elif isinstance(layer, dict):
                    self._layers.append(
                        ManagedLayer(str(layer["name"]), layer, _readonly=_readonly)
                    )
                else:
                    raise TypeError

    def index(self, k):
        for i, u in enumerate(self._layers):
            if u.name == k:
                return i
        return -1

    def __contains__(self, k):
        return self.index(k) != -1

    def __getitem__(self, k):
        """Indexes into the list of layers by index, slice, or layer name."""
        if isinstance(k, str):
            return self._layers[self.index(k)]
        return self._layers[k]

    def __setitem__(self, k, v):
        if self._readonly:
            raise AttributeError
        if isinstance(k, str):
            i = self.index(k)
            if isinstance(v, Layer):
                v = ManagedLayer(k, v)
            elif not isinstance(v, ManagedLayer):
                raise TypeError
            if i == -1:
                self._layers.append(v)
            else:
                self._layers[i] = v
        else:
            if isinstance(k, slice):
                values = []
                for x in v:
                    if not isinstance(v, ManagedLayer):
                        raise TypeError
                    values.append(x)
                self._layers[k] = values
            else:
                if not isinstance(v, ManagedLayer):
                    raise TypeError
                self._layers[k] = v

    def clear(self):
        """Clears the list of layers."""
        del self[:]

    def __delitem__(self, k):
        """Deletes a layer by index, slice, or name."""
        if self._readonly:
            raise AttributeError
        if isinstance(k, str):
            k = self.index(k)
        del self._layers[k]

    def append(self, *args, **kwargs):
        """Appends a ManagedLayer to the list of layers."""
        if self._readonly:
            raise AttributeError
        if len(args) == 1 and not kwargs and isinstance(args[0], ManagedLayer):
            layer = args[0]
        else:
            layer = ManagedLayer(*args, **kwargs)
        self._layers.append(layer)

    def extend(self, elements):
        for element in elements:
            self.append(element)

    def __len__(self):
        """Returns the number of layers in the list."""
        return len(self._layers)

    def __iter__(self):
        return iter(self._layers)

    def to_json(self):
        r = []
        for x in self._layers:
            r.append(x.to_json())
        return r

    def __repr__(self):
        return repr(self._layers)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for layer in c:
            index = b.index(layer.name)
            if index == -1:
                continue
            other_layer = b[index]
            if type(other_layer.layer) is not type(layer.layer):  # noqa: E721
                continue
            layer.layer = type(layer.layer).interpolate(
                layer.layer, other_layer.layer, t
            )
        return c


def navigation_link_type(x):
    x = str(x)
    x = x.lower()
    if x not in ["linked", "unlinked", "relative"]:
        raise ValueError("Invalid navigation link type: %r" % x)
    return x


_set_type_annotation(
    navigation_link_type, typing.Literal["linked", "unlinked", "relative"]
)


@export
class LinkedType(typing.Generic[T], JsonObjectWrapper):
    """Value linked to another value in the viewer state.

    Type parameters:
      T: Value type.
    """

    __slots__ = ()
    link = wrapped_property("link", optional(navigation_link_type, "linked"))
    value: T

    _interpolate_function: typing.ClassVar[
        typing.Callable[[typing.Any, typing.Any, float], typing.Any]
    ]

    @classmethod
    def interpolate(cls, a, b, t):
        c = copy.deepcopy(a)
        c.link = a.link
        if a.link == b.link and a.link != "linked":
            c.value = cls._interpolate_function(a.value, b.value, t)
            return c
        return c


def make_linked_navigation_type(
    value_type: typing.Callable[[typing.Any], T], interpolate_function=None
) -> type[LinkedType[T]]:
    if interpolate_function is None:
        interpolate_function = value_type.interpolate  # type: ignore[attr-defined]

    class Linked(LinkedType):
        __slots__ = ()
        _value_type = value_type
        _interpolate_function = interpolate_function
        value = wrapped_property("value", optional(value_type))

    return Linked


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LinkedPositionBase = LinkedType[np.typing.NDArray[np.float32]]
else:
    _LinkedPositionBase = make_linked_navigation_type(
        array_wrapper(np.float32), interpolate_linear_optional_vectors
    )


@export
class LinkedPosition(_LinkedPositionBase):
    __slots__ = ()


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LinkedZoomFactorBase = LinkedType[float]
else:
    _LinkedZoomFactorBase = make_linked_navigation_type(float, interpolate_zoom)


@export
class LinkedZoomFactor(_LinkedZoomFactorBase):
    __slots__ = ()


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LinkedDepthRangeBase = LinkedType[float]
else:
    _LinkedDepthRangeBase = make_linked_navigation_type(float, interpolate_zoom)


@export
class LinkedDepthRange(_LinkedDepthRangeBase):
    __slots__ = ()


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _LinkedOrientationStateBase = LinkedType[np.typing.NDArray[np.float32]]
else:
    _LinkedOrientationStateBase = make_linked_navigation_type(
        array_wrapper(np.float32, 4), quaternion_slerp
    )


@export
class LinkedOrientationState(_LinkedOrientationStateBase):
    __slots__ = ()


@export
class CrossSection(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    width = wrapped_property("width", optional(int, 1000))
    height = wrapped_property("height", optional(int, 1000))
    position = wrapped_property("position", LinkedPosition)
    orientation = wrapped_property("orientation", LinkedOrientationState)
    scale = wrapped_property("scale", LinkedZoomFactor)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.width = interpolate_linear(a.width, b.width, t)
        c.height = interpolate_linear(a.height, b.height, t)
        c.position = LinkedPosition.interpolate(a.position, b.position, t)
        c.orientation = LinkedOrientationState.interpolate(
            a.orientation, b.orientation, t
        )
        c.scale = LinkedZoomFactor.interpolate(a.scale, b.scale, t)
        return c


if typing.TYPE_CHECKING or _BUILDING_DOCS:
    _CrossSectionMapBase = Map[str, CrossSection]
else:
    _CrossSectionMapBase = typed_map(str, CrossSection)


@export
class CrossSectionMap(_CrossSectionMapBase):
    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in a:
            if k in b:
                c[k] = CrossSection.interpolate(a[k], b[k], t)
        return c


@export
class DataPanelLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", str)
    cross_sections = crossSections = wrapped_property("crossSections", CrossSectionMap)
    orthographic_projection = orthographicProjection = wrapped_property(
        "orthographicProjection", optional(bool, False)
    )

    def __init__(self, json_data=None, _readonly=False, **kwargs):
        if isinstance(json_data, str):
            json_data = {"type": str(json_data)}
        super().__init__(json_data, _readonly=_readonly, **kwargs)

    def to_json(self):
        if len(self.cross_sections) == 0 and not self.orthographic_projection:
            return self.type
        return super().to_json()

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.cross_sections) == 0:
            return a
        c = copy.deepcopy(a)
        c.cross_sections = CrossSectionMap.interpolate(
            a.cross_sections, b.cross_sections, t
        )
        return c


def data_panel_layout_wrapper(default_value="xy"):
    def wrapper(x, _readonly=False):
        if x is None:
            x = default_value
        if isinstance(x, str):
            x = {"type": str(x)}
        return DataPanelLayout(x, _readonly=_readonly)

    wrapper.supports_readonly = True
    return wrapper


data_panel_layout_types = frozenset(
    ["xy", "yz", "xz", "xy-3d", "yz-3d", "xz-3d", "4panel", "4panel-alt", "3d"]
)


def layout_specification(x, _readonly=False):
    if x is None:
        x = "4panel"
    if isinstance(x, str):
        x = {"type": str(x)}
    if isinstance(x, StackLayout | LayerGroupViewer | DataPanelLayout):
        return type(x)(x.to_json(), _readonly=_readonly)
    if not isinstance(x, dict):
        raise ValueError
    layout_type = layout_types.get(x.get("type"))
    if layout_type is None:
        raise ValueError
    return layout_type(x, _readonly=_readonly)


_set_type_annotation(
    layout_specification,
    typing.Union["StackLayout", "LayerGroupViewer", "DataPanelLayout"],
)


layout_specification.supports_readonly = True  # type: ignore[attr-defined]


@export
class StackLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", str)
    flex = wrapped_property("flex", optional(float, 1))
    children = wrapped_property("children", typed_list(layout_specification))

    def __getitem__(self, key):
        return self.children[key]

    def __len__(self):
        return len(self.children)

    def __setitem__(self, key, value):
        self.children[key] = value

    def __delitem__(self, key):
        del self.children[key]

    def __iter__(self):
        return iter(self.children)

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.children) != len(b.children):
            return a
        c = copy.deepcopy(a)
        c.children = [
            interpolate_layout(a_child, b_child, t)
            for a_child, b_child in zip(a.children, b.children)
        ]
        return c


@export
def row_layout(children):
    """Creates a row-oriented `StackLayout`.

    Group:
      viewer-state
    """
    return StackLayout(type="row", children=children)


@export
def column_layout(children):
    """Creates a column-oriented `StackLayout`.

    Group:
      viewer-state
    """
    return StackLayout(type="column", children=children)


def interpolate_layout(a, b, t):
    if type(a) is not type(b):
        return a
    return type(a).interpolate(a, b, t)


@export
class LayerGroupViewer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", str)
    flex = wrapped_property("flex", optional(float, 1))
    layers = wrapped_property("layers", typed_list(str))
    layout = wrapped_property("layout", data_panel_layout_wrapper("xy"))
    position = wrapped_property("position", LinkedPosition)
    velocity = wrapped_property(
        "velocity", typed_map(key_type=str, value_type=DimensionPlaybackVelocity)
    )
    cross_section_orientation = crossSectionOrientation = wrapped_property(
        "crossSectionOrientation", LinkedOrientationState
    )
    cross_section_scale = crossSectionScale = wrapped_property(
        "crossSectionScale", LinkedZoomFactor
    )
    cross_section_depth = crossSectionDepth = wrapped_property(
        "crossSectionDepth", LinkedDepthRange
    )
    projection_orientation = projectionOrientation = wrapped_property(
        "projectionOrientation", LinkedOrientationState
    )
    projection_scale = projectionScale = wrapped_property(
        "projectionScale", LinkedZoomFactor
    )
    projection_depth = projectionDepth = wrapped_property(
        "projectionDepth", LinkedDepthRange
    )
    tool_bindings = toolBindings = wrapped_property(
        "toolBindings", typed_map(key_type=str, value_type=Tool)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = "viewer"

    def __repr__(self):
        j = self.to_json()
        j.pop("type", None)
        return f"{type(self).__name__}({encode_json_for_repr(j)})"

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in (
            "layout",
            "position",
            "cross_section_orientation",
            "cross_section_zoom",
            "perspective_orientation",
            "perspective_zoom",
        ):
            a_attr = getattr(a, k)
            b_attr = getattr(b, k)
            setattr(c, k, type(a_attr).interpolate(a_attr, b_attr, t))
        return c


layout_types = {
    "row": StackLayout,
    "column": StackLayout,
    "viewer": LayerGroupViewer,
}


def add_data_panel_layout_types():
    for k in data_panel_layout_types:
        layout_types[k] = DataPanelLayout


add_data_panel_layout_types()


@export
class ViewerState(JsonObjectWrapper):
    __slots__ = ()
    title = wrapped_property("title", optional(str))
    dimensions = wrapped_property("dimensions", CoordinateSpace)
    relative_display_scales = relativeDisplayScales = wrapped_property(
        "relativeDisplayScales", optional(typed_map(str, float))
    )
    display_dimensions = displayDimensions = wrapped_property(
        "displayDimensions", optional(typed_list(str))
    )
    position = voxel_coordinates = wrapped_property(
        "position", optional(array_wrapper(np.float32))
    )
    velocity = wrapped_property(
        "velocity", typed_map(key_type=str, value_type=DimensionPlaybackVelocity)
    )
    cross_section_orientation = crossSectionOrientation = wrapped_property(
        "crossSectionOrientation", optional(array_wrapper(np.float32, 4))
    )
    cross_section_scale = crossSectionScale = wrapped_property(
        "crossSectionScale", optional(float)
    )
    cross_section_depth = crossSectionDepth = wrapped_property(
        "crossSectionDepth", optional(float)
    )
    projection_scale = projectionScale = wrapped_property(
        "projectionScale", optional(float)
    )
    projection_depth = projectionDepth = wrapped_property(
        "projectionDepth", optional(float)
    )
    projection_orientation = projectionOrientation = perspectiveOrientation = (
        perspective_orientation
    ) = wrapped_property(
        "projectionOrientation", optional(array_wrapper(np.float32, 4))
    )
    show_slices = showSlices = wrapped_property("showSlices", optional(bool, True))
    hide_cross_section_background_3d = hideCrossSectionBackground3D = wrapped_property(
        "hideCrossSectionBackground3D", optional(bool, False)
    )
    show_axis_lines = showAxisLines = wrapped_property(
        "showAxisLines", optional(bool, True)
    )
    wire_frame = wireFrame = wrapped_property("wireFrame", optional(bool, False))
    enable_adaptive_downsampling = enableAdaptiveDownsampling = wrapped_property(
        "enableAdaptiveDownsampling", optional(bool, True)
    )
    show_scale_bar = showScaleBar = wrapped_property(
        "showScaleBar", optional(bool, True)
    )
    show_default_annotations = showDefaultAnnotations = wrapped_property(
        "showDefaultAnnotations", optional(bool, True)
    )
    gpu_memory_limit = gpuMemoryLimit = wrapped_property(
        "gpuMemoryLimit", optional(int)
    )
    system_memory_limit = systemMemoryLimit = wrapped_property(
        "systemMemoryLimit", optional(int)
    )
    concurrent_downloads = concurrentDownloads = wrapped_property(
        "concurrentDownloads", optional(int)
    )
    prefetch = wrapped_property("prefetch", optional(bool, True))
    layers = wrapped_property("layers", Layers)
    layout = wrapped_property("layout", layout_specification)
    cross_section_background_color = crossSectionBackgroundColor = wrapped_property(
        "crossSectionBackgroundColor", optional(str)
    )
    projection_background_color = projectionBackgroundColor = wrapped_property(
        "projectionBackgroundColor", optional(str)
    )
    selected_layer = selectedLayer = wrapped_property(
        "selectedLayer", SelectedLayerState
    )
    statistics = wrapped_property("statistics", StatisticsDisplayState)
    help_panel = helpPanel = wrapped_property("helpPanel", HelpPanelState)
    layer_list_panel = layerListPanel = wrapped_property(
        "layerListPanel", LayerListPanelState
    )
    partial_viewport = partialViewport = wrapped_property(
        "partialViewport",
        optional(
            array_wrapper(np.float64, 4), np.array([0, 0, 1, 1], dtype=np.float64)
        ),
    )
    tool_bindings = toolBindings = wrapped_property(
        "toolBindings", typed_map(key_type=str, value_type=Tool)
    )
    tool_palettes = toolPalettes = wrapped_property(
        "toolPalettes", typed_map(key_type=str, value_type=ToolPalette)
    )

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.position = interpolate_linear_optional_vectors(a.position, b.position, t)
        c.projection_scale = interpolate_zoom(a.projection_scale, b.projection_scale, t)
        c.projection_orientation = quaternion_slerp(
            a.projection_orientation, b.projection_orientation, t
        )
        c.cross_section_scale = interpolate_zoom(
            a.cross_section_scale, b.cross_section_scale, t
        )
        c.cross_section_orientation = quaternion_slerp(
            a.cross_section_orientation, b.cross_section_orientation, t
        )
        c.layers = Layers.interpolate(a.layers, b.layers, t)
        c.layout = interpolate_layout(a.layout, b.layout, t)
        return c



"""
Functions to generate NeuroGlancer shaders.

NeuroGlancer shaders must be snippets of GLSL code. This module contains
functions that dynamically generate such code snippets.

GLSL = OpenGL Shading Language.
"""
# stdlib
import math
import re
from os import PathLike
from textwrap import dedent
from typing import IO

# externals
import numpy as np

# internals
from ngtools import cmdata
from ngtools.local.termcolors import bformat
from ngtools.opener import open


def _flatten(x: list[list]) -> tuple:
    """Transform a list of list into a flat tuple."""
    return tuple(z for y in x for z in y)


def pretty_colormap_list(linewidth: int = 79, colorbar: bool = True) -> str:
    """List all existing colormaps."""
    names = list(filter(lambda x: not x[0] == '_', dir(cmdata)))
    names = ['greyscale', 'orientation'] + names
    names = list(sorted(names))
    longest_name = max(map(len, names))

    if colorbar:
        colors = []
        for name in names:
            if name == "greyscale":
                name = "blackwhite"
            if hasattr(cmdata, name):
                if not isinstance(getattr(cmdata, name), (list, tuple, dict)):
                    colorstr = " " * longest_name
                else:
                    colormap = pycolormaps.make_colormap(name)
                    colorstr = ""
                    for i in range(longest_name):
                        i = i / (longest_name - 1)
                        color = [255 * c for c in colormap(i)]
                        colorstr += bformat.bg.rgb256(*color)(" ")
            else:
                colorstr = " " * longest_name
            colors += [colorstr]

    nbcol = max(1, linewidth // (longest_name + 2))
    nbrow = int(math.ceil(len(names) / nbcol))
    cell = " {:<" + str(longest_name) + "s}"
    row = cell * nbcol
    lastrow = cell * (len(names) % nbcol)
    rep = 2 if colorbar else 1
    if lastrow:
        rows = "\n".join([row] * (nbrow-1) * rep + [lastrow] * rep)
    else:
        rows = "\n".join([row] * nbrow * rep)

    if colorbar:
        cells = []
        for i in range(nbrow):
            cells += names[nbcol*i:nbcol*(i+1)]
            cells += colors[nbcol*i:nbcol*(i+1)]
    else:
        cells = names
    return rows.format(*cells)


def load_fs_lut(
    path: str | PathLike | IO
) -> dict[int, tuple[str, list[float]]]:
    """
    Load a FreeSurfer lookup table.

    Returns a dictionary that maps integer labels to tuples that contains
    1. the structure name, as a string
    2. the structure RGBA color, as a 4-tuple of values between 0 and 1.
    """
    if not hasattr(path, 'readlines'):
        with open(path, 'rt') as f:
            return load_fs_lut(f)
    lut = {}
    for line in path.readlines():
        line = line.split('#')[0].strip()
        if not line:
            continue
        label, name, r, g, b, a, *_ = line.split()
        label, r, g, b, a = int(label), int(r), int(g), int(b), int(a)
        lut[label] = (name, (r/255, g/255, b/255, (a or 255)/255))
    return lut


def rotate_shader(shader: str, mat: list, compose: bool = True) -> str:
    """Apply a rotation matrix to orientation vector of a shader."""
    # No matrix -> return as is
    if mat is None:
        return shader

    # No rotation section -> return as is
    if "<!-- BEGIN ROTATION -->" not in shader:
        return shader

    if hasattr(mat, "tolist"):
        # numpy to list (row major)
        mat = mat.tolist()
    if isinstance(mat[0], (list, tuple)):
        # nested list to flat list (column major)
        ndim = len(mat)
        mat = [mat[i][j] for j in range(ndim) for i in range(ndim)]

    # split sections
    shader_top, shader = shader.split("<!-- BEGIN ROTATION -->")
    shader_mid, shader_btm = shader.split("<!-- END ROTATION -->")
    shader_mid = shader_mid.strip().split("\n")
    # remove comments
    shader_mid = [line.split("//")[0].strip() for line in shader_mid]
    # remove empty lines
    shader_mid = [line for line in shader_mid if line]
    # join
    shader_mid = "\n".join(shader_mid)
    # regex
    pattern = (
        r"\s*mat3\s+(?P<var>[^\s=]+)\s*=\s*mat3\s*\("
        r"\s*(?P<M00>[^\s,]+)\s*,"
        r"\s*(?P<M10>[^\s,]+)\s*,"
        r"\s*(?P<M20>[^\s,]+)\s*,"
        r"\s*(?P<M01>[^\s,]+)\s*,"
        r"\s*(?P<M11>[^\s,]+)\s*,"
        r"\s*(?P<M21>[^\s,]+)\s*,"
        r"\s*(?P<M02>[^\s,]+)\s*,"
        r"\s*(?P<M12>[^\s,]+)\s*,"
        r"\s*(?P<M22>[^\s\)]+)\s*\)"
        r"\s*;\s*"
    )
    match = re.fullmatch(pattern, shader_mid)
    if not match:
        raise ValueError("Cannot parse 3x3 matrix")
    match = {
        key: float(val) if key != "var" else val
        for key, val in match.groupdict().items()
    }

    if compose:
        mat0 = np.asarray([
            [match["M00"], match["M01"], match["M02"]],
            [match["M10"], match["M11"], match["M12"]],
            [match["M20"], match["M21"], match["M22"]],
        ])
        mat1 = np.asarray([
            [mat[0], mat[3], mat[6]],
            [mat[1], mat[4], mat[7]],
            [mat[2], mat[5], mat[8]],
        ])
        mat = mat1 @ mat0
        # to flat column-major
        mat = mat.tolist()
        mat = [mat[i][j] for j in range(ndim) for i in range(ndim)]

    shader_mid = dedent(
        f"""
        mat3 {match["var"]} = mat3(
            {mat[0]}, {mat[1]}, {mat[2]},
            {mat[3]}, {mat[4]}, {mat[5]},
            {mat[6]}, {mat[7]}, {mat[8]}
        );
        """
    )

    # put it back together
    shader_top = shader_top.rstrip()
    if shader_top[-2:] == "//":
        shader_top = shader_top[:-2].rstrip()
    shader_btm = shader_btm.lstrip()

    shader = (
        shader_top +
        "\n// <!-- BEGIN ROTATION -->\n" +
        shader_mid.strip() +
        "\n// <!-- END ROTATION -->\n" +
        shader_btm
    )
    return shader


class pycolormaps:
    """
    Namespace for dynamic colormaps.

    Functions in this class return python functions.
    """

    _DEFAULT_LENGTH = object()

    @staticmethod
    def make_colormap(
        name: str,
        data: object | None = None,
    ) -> str:
        """Generate GLSL code for a colormap."""
        data = data or getattr(cmdata, name)
        if isinstance(data, list):
            return pycolormaps.make_listed(name, data)
        elif isinstance(data, tuple):
            return pycolormaps.make_segmented(name, data)
        elif isinstance(data, dict):
            return pycolormaps.make_weighted(name, data)
        return None

    @staticmethod
    def make_listed(
        name: str,
        data: object | None = None,
    ) -> str:
        """Generate GLSL code for a listed colormap."""
        data = data or getattr(cmdata, name)
        n = len(data)
        data = _flatten(data)

        def eval_color(x: float) -> list[float]:
            y = x * (n - 1)
            i = int(math.floor(y))
            j = min(i + 1, n-1)
            w = y - i
            return [
                (1.0-w) * data[3*i+0] + w * data[3*j+0],
                (1.0-w) * data[3*i+1] + w * data[3*j+1],
                (1.0-w) * data[3*i+2] + w * data[3*j+2],
            ]

        return eval_color

    @staticmethod
    def make_weighted(
        name: str,
        data: object | None = None,
    ) -> str:
        """Generate GLSL code for a listed colormap."""
        data = data or getattr(cmdata, name)
        keys = list(data.keys())

        def eval_color(x: float) -> list[float]:
            wj, dj = keys[0], next(iter(data.values()))
            for wi, di in data.items():
                if x <= wi:
                    break
                wj, dj = wi, di
            wlow = x - wi
            wupp = wj - x
            wsum = wlow + wupp
            if wsum == 0:
                wlow = wupp = 0.5
            else:
                wlow = wlow / wsum
                wupp = wupp / wsum
            return [
                di[0] * wupp + dj[0] * wlow,
                di[1] * wupp + dj[1] * wlow,
                di[2] * wupp + dj[2] * wlow,
            ]

        return eval_color

    @staticmethod
    def make_segmented(
        name: str,
        data: object | None = None,
    ) -> str:
        """Generate GLSL code for a segmented colormap."""
        r, g, b = data or getattr(cmdata, name)
        r, g, b = map(_flatten, (r, g, b))

        def segment(x: float, data: list[float]) -> float:
            j = 0
            for i in range(len(data)//3):
                if x <= data[3*i]:
                    break
                j = i
            wlow = x - data[3*i]
            wupp = data[3*j] - x
            wsum = wlow + wupp
            if wsum == 0:
                wlow = wupp = 0.5
            else:
                wlow = wlow / wsum
                wupp = wupp / wsum
            x = data[3*i+2] * wupp + data[3*j+1] * wlow
            return min(max(x, 0), 1)

        def eval_color(x: float) -> list[float]:
            return [segment(x, r), segment(x, g), segment(x, b)]

        return eval_color


class colormaps:
    """
    Namespace for dynamic colormaps.

    Functions in this class return shader code.
    """

    _DEFAULT_LENGTH = object()

    orientation = dedent(
        """
        vec3 colormapOrient(vec3 orient) {
        vec3 result;
        result.r = abs(orient[0]);
        result.g = abs(orient[1]);
        result.b = abs(orient[2]);
        return clamp(result, 0.0, 1.0);
        }
        """).lstrip()

    @staticmethod
    def make_colormap(
        name: str,
        data: object | None = None,
        n: int = _DEFAULT_LENGTH,
    ) -> str:
        """Generate GLSL code for a colormap."""
        data = data or getattr(cmdata, name)
        if isinstance(data, list):
            return colormaps.make_listed(name, data, n)
        elif isinstance(data, tuple):
            return colormaps.make_segmented(name, data, n)
        elif isinstance(data, dict):
            return None
        return None

    @staticmethod
    def make_listed(
        name: str,
        data: object | None = None,
        n: int = _DEFAULT_LENGTH,
    ) -> str:
        """Generate GLSL code for a listed colormap."""
        data = data or getattr(cmdata, name)
        if n is colormaps._DEFAULT_LENGTH:
            n = 128
        if n:
            step = max(1, len(data) // n)
            data = data[::step]
        data = _flatten(data)
        n = len(data)
        return dedent(
            """
            vec4 %s(float x, bool alpha) {
                float cmap[%d] = float[]%s;
                float y = x * float(%d/3 - 1);
                int   i = int(floor(y));
                int   j = i + 1;
                float w = y - float(i);
                vec4  result;
                result.r = (1.0-w) * cmap[3*i+0] + w * cmap[3*j+0];
                result.g = (1.0-w) * cmap[3*i+1] + w * cmap[3*j+1];
                result.b = (1.0-w) * cmap[3*i+2] + w * cmap[3*j+2];
                result.a = (alpha ? x : 1.0);
                return clamp(result, 0.0, 1.0);
            }
            """ % (name, n, str(data), n)
        ).strip()

    @staticmethod
    def make_segmented(
        name: str,
        data: object | None = None,
        n: int = _DEFAULT_LENGTH,
    ) -> str:
        """Generate GLSL code for a segmented colormap."""
        def subsample(data: list[float], n: int) -> list[float]:
            if n is None:
                return cmap
            step = max(1, len(r) // n)
            if len(data) % step != 1:
                last = data[-1:]
            else:
                last = []
            data = data[::step] + last
            return data

        if n is colormaps._DEFAULT_LENGTH:
            n = 32

        r, g, b = data or getattr(cmdata, name)
        r, g, b = map(lambda x: subsample(x, n), (r, g, b))
        r, g, b = map(_flatten, (r, g, b))
        nr, ng, nb = len(r), len(g), len(b)
        segment = dedent(
            """
            float segment%s(float x)
            {
                float cmap[%d] = float[]%s;
                int i, j;
                for (int c = 0; c < cmap.length()/3; ++c) {
                    if (x > cmap[3*c]) {
                        i = c;
                        break;
                    }
                }
                j = (i+1 < cmap.length() ? i+1 : i);
                float wlow = x - cmap[3*i];
                float wupp = cmap[3*j] - x;
                float wsum = wlow + wupp;
                wlow = wlow / wsum;
                wupp = wupp / wsum;
                x = cmap[3*i+2] * wupp + cmap[3*j+1] * wlow;
                return clamp(x, 0.0, 1.0);
            }
            """
        ).strip()
        segment = [segment % (name, length, str(cmap)) for name, length, cmap
                   in (('r', nr, r), ('g', ng, g), ('b', nb, b))]
        segment = '\n'.join(segment)
        return segment + dedent(
            """
            vec4 %s(float x, bool alpha) {
                vec4  result;
                result.r = segmentr(x);
                result.g = segmentg(x);
                result.b = segmentb(x);
                result.a = (alpha ? x : 1.0);
                return clamp(result, 0.0, 1.0);
            }
            """ % name
        ).strip()


for name in filter(lambda x: not x[0] == '_', dir(cmdata)):
    cmap = colormaps.make_colormap(name)
    if cmap:
        setattr(colormaps, name, cmap)


class shaders:
    """Namespace for full-fledged shaders."""

    class skeleton:
        """Shaders for skeletons."""

        # Shader for streamlines that have an `orientation` attribute
        orientation = colormaps.orientation + '\n' + dedent(
            """
            #uicontrol bool orient_color checkbox(default=true)
            void main() {
            vec3 orient = orientation;
            // <!-- BEGIN ROTATION -->
            // Order: 00 10 20 01 11 21 02 12 22
            mat3 mat = mat3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
            // <!-- END ROTATION -->
            orient = mat * orient;
            if (orient_color)
                emitRGB(colormapOrient(orient));
            else
                emitDefault();
            }
            """).lstrip()

    default = dedent(
        """
        #uicontrol invlerp normalized
        void main() {
        emitGrayscale(normalized());
        }
        """
    )

    # Classic greyscale map, with optional intensity-based opacity.
    greyscale = dedent(
        """
        #uicontrol invlerp normalized
        #uicontrol bool alpha_depth checkbox(default=false)
        void main() {
            float x = normalized();
            if(alpha_depth) {
                vec4 result;
                result.r = x;
                result.g = x;
                result.b = x;
                result.a = x;
                emitRGBA(result);
            } else {
                emitGrayscale(x);
            }
        }
        """).lstrip()

    # RGB orientation, for vector fields
    orientation = dedent(
        """
        #uicontrol float brightness_ceil slider(min=0, max=1, default=1)
        #uicontrol float brightness_floor slider(min=0, max=1, default=0)
        #uicontrol float alpha_ceil slider(min=0, max=1, default=1)
        #uicontrol float alpha_floor slider(min=0, max=1, default=1)
        void main() {
        vec3 orient;
        orient.r = toNormalized(getDataValue(0));
        orient.g = toNormalized(getDataValue(1));
        orient.b = toNormalized(getDataValue(2));
        // <!-- BEGIN ROTATION -->
        // Order: 00 10 20 01 11 21 02 12 22
        mat3 mat = mat3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
        // <!-- END ROTATION -->
        orient = mat * orient;
        vec4 rgba;
        rgba.r = abs(orient.r);
        rgba.g = abs(orient.g);
        rgba.b = abs(orient.b);
        rgba.a = sqrt(
            rgba.r * rgba.r +
            rgba.g * rgba.g +
            rgba.b * rgba.b
        );
        rgba.r /= rgba.a;
        rgba.g /= rgba.a;
        rgba.b /= rgba.a;

        float brightness = brightness_ceil * (
            (1.0 - brightness_floor) * rgba.a + brightness_floor
        );
        rgba.r *= brightness;
        rgba.g *= brightness;
        rgba.b *= brightness;
        rgba.a *= alpha_ceil * (1.0 - alpha_floor);
        rgba.a += alpha_ceil * alpha_floor;

        emitRGBA(rgba);
        }
        """
    )

    # Classic RGB map, with brightness + contrast controls
    rgb = dedent(
        """
        #uicontrol float brightness slider(min=-1, max=1)
        #uicontrol float contrast slider(min=-3, max=3, step=0.01)
        void main() {
            vec3 color;
            color.r = toNormalized(getDataValue(0));
            color.g = toNormalized(getDataValue(1));
            color.b = toNormalized(getDataValue(2));
            emitRGB((color + brightness) * exp(contrast));
        }
        """
    )

    @staticmethod
    def colormap(cmap: str) -> str:
        """
        Generate a shader based on a known colormap.

        Parameters
        ----------
        cmap : str
            Name of a colormap. Must be a valid attribute from the
            `colormaps` namespace.

        """
        return getattr(colormaps, cmap) + '\n' + dedent(
            """
            #uicontrol invlerp normalized
            #uicontrol bool alpha_depth checkbox(default=false)
            void main() {
                emitRGBA(%s(normalized(), alpha_depth));
            }
            """ % cmap).lstrip()

    @staticmethod
    def lut(lut: dict | str | PathLike | IO) -> str:
        """
        Generate a shader based on a lookup table.

        Parameters
        ----------
        lut : str | dict[int, tuple[str, tuple[float,float,float,float]]]
            Path to a FreeSurfer lookup table, or pre-loaded LUT.
        """
        if not isinstance(lut, dict):
            lut = load_fs_lut(lut)
        labels, colors = list(lut.keys()), list(lut.values())
        colors = [color[1] for color in colors]
        shader = dedent(
            """
            void main() {
            vec4 color;
            int label = int(getDataValue(0).value);
            if (label == int(%d))
                color = vec4(%f, %f, %f, %f);
            """ % (labels.pop(0), *colors.pop(0))
        )
        for label, color in zip(labels, colors):
            shader += dedent(
                """
                else if (label == int(%d))
                    color = vec4(%f, %f, %f, %f);
                """ % (label, *color)
            )
        shader += dedent(
            """
            else
                color = vec4(0.0, 0.0, 0.0, 0.0);
            emitRGBA(color);
            }
            """
        )
        return shader



"""Handlers for the WSGI server."""
# stdlib
import json
import logging

# externals
import requests

# internals
from ngtools.local.fileserver import Handler
from ngtools.opener import exists, linc_auth_opt
from ngtools.shaders import load_fs_lut

LOG = logging.getLogger(__name__)


class LutHandler(Handler):
    """Handler that returns a segment_properties from a FS LUT."""

    def get(self, protocol: str, path: str) -> None:  # noqa: D102

        if not path.endswith("info"):
            return self.send_error(404, path)

        path = path[:-5]
        path = protocol + "://" + path

        if not exists(path):
            return self.send_error(404, path)

        try:
            lut = load_fs_lut(path)
        except Exception as e:
            return self.send_error(400, e)

        segment_properties = {
            "@type": "neuroglancer_segment_properties",
            "inline": {
                "ids": list(map(str, lut.keys())),
                "properties": [
                    {
                        "id": "name",
                        "type": "label",
                        "description": "Name",
                        "values": [name for name, _ in lut.values()]
                    }
                ]
            }
        }

        self.status = 200
        self.headers["Content-type"] = "application/json"
        self.body = json.dumps(segment_properties).encode()


class LincHandler(Handler):
    """Handler that redirects to linc data."""

    VALID_KEYS = (
        "Accept",
        "Accept-encoding",
        "Content-type",
        "Content-length",
        # "Connection",  # hop-by-hop
        "Last-modified",
        "Accept-ranges",
        "Range",
        "Date",
        "Etag",
        "Vary",
    )

    # NOTE: Connection is a hop-by-hop header and should therfore not
    # be forwarded to the next request.
    # https://0xn3va.gitbook.io/cheat-sheets/web-application/abusing-http-hop-by-hop-request-headers

    def prepare(self) -> None:  # noqa: D102
        if getattr(self.app, "_linc_cookies", None) is None:
            self.app._linc_cookies = linc_auth_opt()
        if getattr(self.app, "_linc_sessions", None) is None:
            self.app._linc_session = requests.Session()
        self._linc_session = self.app._linc_session
        self._linc_cookies = self.app._linc_cookies

    def _request(self, method: str, path: str) -> None:
        LOG.debug(
            f"LincHandler: {method} {path} << {dict(self.environ.headers)}"
        )
        session = self._linc_session
        cookies = self._linc_cookies

        # select input headers to forward
        headers = {
            key: value
            for key, value in self.environ.headers.items()
            if key.capitalize() in self.VALID_KEYS
        }

        # fetch
        url = "https://neuroglancer.lincbrain.org/" + str(path)
        try:
            fetch = getattr(session, method.lower())
            response = fetch(url, headers=headers, **cookies)

        except Exception as e:
            LOG.debug(f"LincHandler: {path} | {e}")
            return self.send_error(400, e)

        else:
            status = getattr(response, "status_code", 400)
            headers = dict(getattr(response, "headers", {}))
            content = getattr(response, "content", None)

            LOG.debug(
                f"LincHandler: {method} {path} >> {status} {headers}"
            )

        self.status = status

        # select output headers to forward
        for key, val in headers.items():
            if key.capitalize() in self.VALID_KEYS:
                self.headers.add_header(key, val)

        if method.upper() == "GET":
            self.body = content

        LOG.debug(f"{method} | {path} | done")

    def get(self, path: str) -> None:  # noqa: D102
        return self._request("GET", path)

    def head(self, path: str) -> None:  # noqa: D102
        return self._request("HEAD", path)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import numbers

import numpy as np

from . import local_volume

min_safe_integer = -9007199254740991
max_safe_integer = 9007199254740991


def json_encoder_default(obj):
    """JSON encoder function that handles some numpy types."""
    if isinstance(obj, numbers.Integral) and (
        obj < min_safe_integer or obj > max_safe_integer
    ):
        return str(obj)
    if isinstance(obj, np.integer):
        return str(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return list(obj)
    elif isinstance(obj, set | frozenset):
        return list(obj)
    raise TypeError


def json_encoder_default_for_repr(obj):
    if isinstance(obj, local_volume.LocalVolume):
        return "<LocalVolume>"
    return json_encoder_default(obj)


def decode_json(x):
    return json.loads(x)


def encode_json(obj):
    return json.dumps(obj, default=json_encoder_default)


def encode_json_for_repr(obj):
    return json.dumps(obj, default=json_encoder_default_for_repr)



import logging
import os
import threading
from configparser import ConfigParser

from . import credentials_provider
from .futures import run_on_new_thread


class BossCredentialsProvider(credentials_provider.CredentialsProvider):
    def __init__(self):
        super().__init__()

        # Make sure logging is initialized.  Does nothing if logging has already
        # been initialized.
        logging.basicConfig()

        self._lock = threading.Lock()
        self._credentials = None

    def set_token(self, token):
        # Token should be a string
        self._credentials = dict(tokenType="Token", accessToken=token)

    def get_new(self):
        def func():
            with self._lock:
                # First, see if user has defined a token using set_token
                if self._credentials is not None:
                    return self._credentials

                # If not, look for config file in intern file location
                config_path = "~/.intern/intern.cfg"
                if os.path.isfile(os.path.expanduser(config_path)):
                    with open(os.path.expanduser(config_path)) as config_file_handle:
                        config_parser = ConfigParser()
                        config_parser.read_file(config_file_handle)
                        # Try Default section first
                        try:
                            self._credentials = config_parser["Default"]["token"]
                            print("Using token from intern config file")
                            return dict(
                                tokenType="Token", accessToken=self._credentials
                            )
                        except Exception:
                            pass
                        # Try Volume Service section second
                        try:
                            self._credentials = config_parser["Volume Service"]["token"]
                            print("Using token from intern config file")
                            return dict(
                                tokenType="Token", accessToken=self._credentials
                            )
                        except Exception:
                            pass

                # Else, use "public"
                print("Accessing Boss data using token 'public'")
                return dict(tokenType="Token", accessToken="public")

        return run_on_new_thread(func)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common interface for transforms."""

from pathlib import Path
import numpy as np
import warnings
from nibabel.loadsave import load as _nbload
from nibabel import funcs as _nbfuncs
from nibabel.nifti1 import intent_codes as INTENT_CODES
from nibabel.cifti2 import Cifti2Image
import nibabel as nb

EQUALITY_TOL = 1e-5


class TransformError(TypeError):
    """A custom exception for transforms."""


class SpatialReference:
    """Factory to create spatial references."""

    @staticmethod
    def factory(dataset):
        """Create a reference for spatial transforms."""
        try:
            return SampledSpatialData(dataset)
        except ValueError:
            return ImageGrid(dataset)


class SampledSpatialData:
    """Represent sampled spatial data: regularly gridded (images) and surfaces."""

    __slots__ = ["_ndim", "_coords", "_npoints", "_shape"]

    def __init__(self, dataset):
        """Create a sampling reference."""
        self._shape = None

        if isinstance(dataset, SampledSpatialData):
            self._coords = dataset.ndcoords.copy()
            self._npoints, self._ndim = self._coords.shape
            return

        if isinstance(dataset, (str, Path)):
            dataset = _nbload(str(dataset))

        if hasattr(dataset, "numDA"):  # Looks like a Gifti file
            _das = dataset.get_arrays_from_intent(INTENT_CODES["pointset"])
            if not _das:
                raise TypeError(
                    "Input Gifti file does not contain reference coordinates."
                )
            self._coords = np.vstack([da.data for da in _das])
            self._npoints, self._ndim = self._coords.shape
            return

        if isinstance(dataset, Cifti2Image):
            raise NotImplementedError

        raise ValueError("Dataset could not be interpreted as an irregular sample.")

    @property
    def npoints(self):
        """Access the total number of voxels."""
        return self._npoints

    @property
    def ndim(self):
        """Access the number of dimensions."""
        return self._ndim

    @property
    def ndcoords(self):
        """List the physical coordinates of this sample."""
        return self._coords

    @property
    def shape(self):
        """Access the space's size of each dimension."""
        return self._shape


class SurfaceMesh(SampledSpatialData):
    """Class to represent surface meshes."""

    __slots__ = ["_triangles"]

    def __init__(self, dataset):
        """Create a sampling reference."""
        self._shape = None

        if isinstance(dataset, SurfaceMesh):
            self._coords = dataset._coords
            self._triangles = dataset._triangles
            self._ndim = dataset._ndim
            self._npoints = dataset._npoints
            self._shape = dataset._shape
            return

        if isinstance(dataset, (str, Path)):
            dataset = _nbload(str(dataset))

        if hasattr(dataset, "numDA"):  # Looks like a Gifti file
            _das = dataset.get_arrays_from_intent(INTENT_CODES["pointset"])
            if not _das:
                raise TypeError(
                    "Input Gifti file does not contain reference coordinates."
                )
            self._coords = np.vstack([da.data for da in _das])
            _tris = dataset.get_arrays_from_intent(INTENT_CODES["triangle"])
            self._triangles = np.vstack([da.data for da in _tris])
            self._npoints, self._ndim = self._coords.shape
            self._shape = self._coords.shape
            return

        if isinstance(dataset, Cifti2Image):
            raise NotImplementedError

        raise ValueError("Dataset could not be interpreted as an irregular sample.")

    def check_sphere(self, tolerance=1.001):
        """Check sphericity of surface.
        Based on https://github.com/Washington-University/workbench/blob/\
7ba3345d161d567a4b628ceb02ab4471fc96cb20/src/Files/SurfaceResamplingHelper.cxx#L503
        """
        dists = np.linalg.norm(self._coords, axis=1)
        return (dists.min() * tolerance) > dists.max()

    def set_radius(self, radius=100):
        if not self.check_sphere():
            raise ValueError("You should only set the radius on spherical surfaces.")
        dists = np.linalg.norm(self._coords, axis=1)
        self._coords = self._coords * (radius / dists).reshape((-1, 1))

    @classmethod
    def from_arrays(cls, coordinates, triangles):
        darrays = [
            nb.gifti.GiftiDataArray(
                coordinates.astype(np.float32),
                intent=nb.nifti1.intent_codes["NIFTI_INTENT_POINTSET"],
                datatype=nb.nifti1.data_type_codes["NIFTI_TYPE_FLOAT32"],
            ),
            nb.gifti.GiftiDataArray(
                triangles.astype(np.int32),
                intent=nb.nifti1.intent_codes["NIFTI_INTENT_TRIANGLE"],
                datatype=nb.nifti1.data_type_codes["NIFTI_TYPE_INT32"],
            ),
        ]
        gii = nb.gifti.GiftiImage(darrays=darrays)
        return cls(gii)


class ImageGrid(SampledSpatialData):
    """Class to represent spaces of gridded data (images)."""

    __slots__ = ["_affine", "_inverse", "_ndindex", "_header"]

    def __init__(self, image):
        """Create a gridded sampling reference."""
        if isinstance(image, (str, Path)):
            image = _nbfuncs.squeeze_image(_nbload(str(image)))

        self._affine = image.affine
        self._shape = image.shape
        self._header = getattr(image, "header", None)

        self._ndim = getattr(image, "ndim", len(image.shape))
        if self._ndim >= 4:
            self._shape = image.shape[:3]
            self._ndim = 3

        self._npoints = getattr(image, "npoints", np.prod(self._shape))
        self._ndindex = None
        self._coords = None
        self._inverse = getattr(image, "inverse", np.linalg.inv(image.affine))

    @property
    def affine(self):
        """Access the indexes-to-RAS affine."""
        return self._affine

    @property
    def header(self):
        """Access the original reference's header."""
        return self._header

    @property
    def inverse(self):
        """Access the RAS-to-indexes affine."""
        return self._inverse

    @property
    def ndindex(self):
        """List the indexes corresponding to the space grid."""
        if self._ndindex is None:
            indexes = tuple([np.arange(s) for s in self._shape])
            self._ndindex = np.array(np.meshgrid(*indexes, indexing="ij")).reshape(
                self._ndim, self._npoints
            )
        return self._ndindex

    @property
    def ndcoords(self):
        """List the physical coordinates of this gridded space samples."""
        if self._coords is None:
            self._coords = np.tensordot(
                self._affine,
                np.vstack((self.ndindex, np.ones((1, self._npoints)))),
                axes=1,
            )[:3, ...]
        return self._coords

    def ras(self, ijk):
        """Get RAS+ coordinates from input indexes."""
        return _apply_affine(ijk, self._affine, self._ndim)

    def index(self, x):
        """Get the image array's indexes corresponding to coordinates."""
        return _apply_affine(x, self._inverse, self._ndim)

    def _to_hdf5(self, group):
        group.attrs["Type"] = "image"
        group.attrs["ndim"] = self.ndim
        group.create_dataset("affine", data=self.affine)
        group.create_dataset("shape", data=self.shape)

    def __eq__(self, other):
        """Overload equals operator."""
        return (
            np.allclose(self.affine, other.affine, rtol=EQUALITY_TOL)
            and self.shape == other.shape
        )

    def __ne__(self, other):
        """Overload not equal operator."""
        return not self == other


class TransformBase:
    """Abstract image class to represent transforms."""

    __slots__ = (
        "_reference",
        "_ndim",
    )

    def __init__(self, reference=None):
        """Instantiate a transform."""
        self._reference = None
        if reference:
            self.reference = reference

    def __call__(self, x, inverse=False):
        """Apply y = f(x)."""
        return self.map(x, inverse=inverse)

    def __add__(self, b):
        """
        Compose this and other transforms.

        Example
        -------
        >>> T1 = TransformBase()
        >>> added = T1 + TransformBase()
        >>> len(added.transforms)
        2

        """
        from .manip import TransformChain

        return TransformChain(transforms=[self, b])

    def __len__(self):
        """
        Enable ``len()``.

        By default, all transforms are of length one.
        This must be overriden by transforms arrays and chains.

        Example
        -------
        >>> T1 = TransformBase()
        >>> len(T1)
        1

        """
        return 1

    @property
    def reference(self):
        """Access a reference space where data will be resampled onto."""
        if self._reference is None:
            warnings.warn("Reference space not set")
        return self._reference

    @reference.setter
    def reference(self, image):
        self._reference = ImageGrid(image)

    @property
    def ndim(self):
        """Access the dimensions of the reference space."""
        raise TypeError("TransformBase has no dimensions")

    def map(self, x, inverse=False):
        r"""
        Apply :math:`y = f(x)`.

        TransformBase implements the identity transform.

        Parameters
        ----------
        x : N x D numpy.ndarray
            Input RAS+ coordinates (i.e., physical coordinates).
        inverse : bool
            If ``True``, apply the inverse transform :math:`x = f^{-1}(y)`.

        Returns
        -------
        y : N x D numpy.ndarray
            Transformed (mapped) RAS+ coordinates (i.e., physical coordinates).

        """
        return x

    def to_filename(self, filename, fmt="X5", moving=None, x5_inverse=False):
        """Store the transform in BIDS-Transforms HDF5 file format (.x5)."""
        raise NotImplementedError

    def _to_hdf5(self, x5_root):
        """Serialize this object into the x5 file format."""
        raise NotImplementedError

    def apply(self, *args, **kwargs):
        """Apply the transform to a dataset.

        Deprecated. Please use ``nitransforms.resampling.apply`` instead.
        """
        _msg = "This method is deprecated. Please use `nitransforms.resampling.apply` instead."
        warnings.warn(_msg, DeprecationWarning, stacklevel=2)
        from .resampling import apply

        return apply(self, *args, **kwargs)


def _as_homogeneous(xyz, dtype="float32", dim=3):
    """
    Convert 2D and 3D coordinates into homogeneous coordinates.

    Examples
    --------
    >>> _as_homogeneous((4, 5), dtype='int8', dim=2).tolist()
    [[4, 5, 1]]

    >>> _as_homogeneous((4, 5, 6),dtype='int8').tolist()
    [[4, 5, 6, 1]]

    >>> _as_homogeneous((4, 5, 6, 1),dtype='int8').tolist()
    [[4, 5, 6, 1]]

    >>> _as_homogeneous([(1, 2, 3), (4, 5, 6)]).tolist()
    [[1.0, 2.0, 3.0, 1.0], [4.0, 5.0, 6.0, 1.0]]


    """
    xyz = np.atleast_2d(np.array(xyz, dtype=dtype))
    if np.shape(xyz)[-1] == dim + 1:
        return xyz

    return np.hstack((xyz, np.ones((xyz.shape[0], 1), dtype=dtype)))


def _apply_affine(x, affine, dim):
    """Get the image array's indexes corresponding to coordinates."""
    return np.tensordot(
        affine,
        _as_homogeneous(x, dim=dim).T,
        axes=1,
    )[:dim, ...]



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Interpolate with 3D tensor-product B-Spline basis."""

import numpy as np
import nibabel as nb
from scipy.sparse import csr_matrix, kron


def _cubic_bspline(d, order=3):
    """Evaluate the cubic B-spline at distance ``d`` from the center."""

    if order != 3:
        raise NotImplementedError

    d = np.abs(d)
    out = np.zeros_like(d, dtype="float32")

    mask1 = d < 1.0
    mask2 = (d >= 1.0) & (d < 2.0)

    out[mask1] = (4.0 - 6.0 * d[mask1] ** 2 + 3.0 * d[mask1] ** 3) / 6.0
    out[mask2] = (2.0 - d[mask2]) ** 3 / 6.0

    return out


def grid_bspline_weights(target_grid, ctrl_grid):
    r"""
    Evaluate tensor-product B-Spline weights on a grid.

    For each of the :math:`N` input locations :math:`\mathbf{x} = (x_i, x_j, x_k)`
    and :math:`K` control points or *knots* :math:`\mathbf{c} =(c_i, c_j, c_k)`,
    the tensor-product cubic B-Spline kernel weights are calculated:

    .. math::
        \Psi^3(\mathbf{x}, \mathbf{c}) =
        \beta^3(x_i - c_i) \cdot \beta^3(x_j - c_j) \cdot \beta^3(x_k - c_k),
        \label{eq:bspline_weights}\tag{1}

    where each :math:`\beta^3` represents the cubic B-Spline for one dimension.
    The 1D B-Spline kernel implementation uses :obj:`numpy.piecewise`, and is based on the
    closed-form given by Eq. (6) of [Unser1999]_.

    By iterating over dimensions, the data samples that fall outside of the compact
    support of the tensor-product kernel associated to each control point can be filtered
    out and dismissed to lighten computation.

    Finally, the resulting weights matrix :math:`\Psi^3(\mathbf{k}, \mathbf{s})`
    can be easily identified in Eq. :math:`\eqref{eq:1}` and used as the design matrix
    for approximation of data.

    Parameters
    ----------
    target_grid : :obj:`~nitransforms.base.ImageGrid` or :obj:`nibabel.spatialimages`
        Regular grid of :math:`N` locations at which tensor B-Spline basis will be evaluated.
    ctrl_grid : :obj:`~nitransforms.base.ImageGrid` or :obj:`nibabel.spatialimages`
        Regular grid of :math:`K` control points (knot) where B-Spline basis are defined.

    Returns
    -------
    weights : :obj:`numpy.ndarray` (:math:`K \times N`)
        A sparse matrix of interpolating weights :math:`\Psi^3(\mathbf{k}, \mathbf{s})`
        for the *N* voxels of the target EPI, for each of the total *K* knots.
        This sparse matrix can be directly used as design matrix for the fitting
        step of approximation/extrapolation.

    """
    shape = target_grid.shape[:3]
    ctrl_sp = nb.affines.voxel_sizes(ctrl_grid.affine)[:3]
    ras2ijk = np.linalg.inv(ctrl_grid.affine)
    # IJK index in the control point image of the first index in the target image
    origin = nb.affines.apply_affine(ras2ijk, [tuple(target_grid.affine[:3, 3])])[0]

    wd = []
    for i, (o, n, sp) in enumerate(
        zip(origin, shape, nb.affines.voxel_sizes(target_grid.affine)[:3])
    ):
        # Locations of voxels in target image in control point image
        locations = np.arange(0, n, dtype="float16") * sp / ctrl_sp[i] + o
        knots = np.arange(0, ctrl_grid.shape[i], dtype="float16")
        distance = np.abs(locations[np.newaxis, ...] - knots[..., np.newaxis])

        within_support = distance < 2.0
        d_vals, d_idxs = np.unique(distance[within_support], return_inverse=True)
        bs_w = _cubic_bspline(d_vals)
        weights = np.zeros_like(distance, dtype="float32")
        weights[within_support] = bs_w[d_idxs]
        wd.append(csr_matrix(weights))

    return kron(kron(wd[0], wd[1]), wd[2])



"""Read/write linear transforms."""

import numpy as np
from nibabel.volumeutils import Recoder
from nibabel.affines import voxel_sizes, from_matvec

from .base import (
    BaseLinearTransformList,
    StringBasedStruct,
    LinearTransformStruct,
    TransformFileError,
)


transform_codes = Recoder(
    (
        (0, "LINEAR_VOX_TO_VOX"),
        (1, "LINEAR_RAS_TO_RAS"),
        (2, "LINEAR_PHYSVOX_TO_PHYSVOX"),
        (14, "REGISTER_DAT"),
        (21, "LINEAR_COR_TO_COR"),
    ),
    fields=("code", "label"),
)


class VolumeGeometry(StringBasedStruct):
    """Data structure for regularly gridded images."""

    template_dtype = np.dtype(
        [
            ("valid", "i4"),  # Valid values: 0, 1
            ("volume", "i4", (3,)),  # width, height, depth
            ("voxelsize", "f4", (3,)),  # xsize, ysize, zsize
            ("xras", "f8", (3, 1)),  # x_r, x_a, x_s
            ("yras", "f8", (3, 1)),  # y_r, y_a, y_s
            ("zras", "f8", (3, 1)),  # z_r, z_a, z_s
            ("cras", "f8", (3,)),  # c_r, c_a, c_s
            ("filename", "U1024"),
        ]
    )  # Not conformant (may be >1024 bytes)
    dtype = template_dtype

    def as_affine(self):
        """Return the internal affine of this regular grid."""
        sa = self.structarr
        A = np.hstack((sa["xras"], sa["yras"], sa["zras"])) * sa["voxelsize"]
        b = sa["cras"] - A @ sa["volume"] / 2
        return from_matvec(A, b)

    def __str__(self):
        """Format the structure as a text file."""
        sa = self.structarr
        lines = [
            "valid = {}  # volume info {:s}valid".format(
                sa["valid"], "" if sa["valid"] else "in"
            ),
            "filename = {}".format(sa["filename"]),
            "volume = {:d} {:d} {:d}".format(*sa["volume"]),
            "voxelsize = {:.15e} {:.15e} {:.15e}".format(*sa["voxelsize"]),
            "xras   = {:.15e} {:.15e} {:.15e}".format(*sa["xras"].flatten()),
            "yras   = {:.15e} {:.15e} {:.15e}".format(*sa["yras"].flatten()),
            "zras   = {:.15e} {:.15e} {:.15e}".format(*sa["zras"].flatten()),
            "cras   = {:.15e} {:.15e} {:.15e}".format(*sa["cras"].flatten()),
        ]
        return "\n".join(lines)

    def to_string(self):
        """Format the structure as a text file."""
        return self.__str__()

    @classmethod
    def from_image(cls, img):
        """Create struct from an image."""
        volgeom = cls()
        sa = volgeom.structarr
        sa["valid"] = 1
        sa["volume"] = img.shape[:3]  # Assumes xyzt-ordered image
        sa["voxelsize"] = voxel_sizes(img.affine)[:3]
        A = img.affine[:3, :3]
        b = img.affine[:3, 3]
        cols = A / sa["voxelsize"]
        sa["xras"] = cols[:, [0]]
        sa["yras"] = cols[:, [1]]
        sa["zras"] = cols[:, [2]]
        sa["cras"] = b + A @ sa["volume"] / 2
        try:
            sa["filename"] = img.file_map["image"].filename
        except Exception:
            pass

        return volgeom

    @classmethod
    def from_string(cls, string):
        """Create a volume structure off of text."""
        volgeom = cls()
        sa = volgeom.structarr
        lines = string.splitlines()
        for key in (
            "valid",
            "filename",
            "volume",
            "voxelsize",
            "xras",
            "yras",
            "zras",
            "cras",
        ):
            label, valstring = lines.pop(0).split(" =")
            assert label.strip() == key

            parsed = (
                np.genfromtxt(
                    [valstring.encode()], autostrip=True, dtype=cls.dtype[key]
                )
                if valstring.strip()
                else None
            )

            if parsed is not None and parsed.size:
                sa[key] = parsed.reshape(sa[key].shape)
            else:  # pragma: no coverage
                """Do not set sa[key]"""

        return volgeom


class FSLinearTransform(LinearTransformStruct):
    """Represents a single LTA's transform structure."""

    template_dtype = np.dtype(
        [
            ("type", "i4"),
            ("mean", "f4", (3, 1)),  # x0, y0, z0
            ("sigma", "f4"),
            ("m_L", "f8", (4, 4)),
            ("m_dL", "f8", (4, 4)),
            ("m_last_dL", "f8", (4, 4)),
            ("src", VolumeGeometry),
            ("dst", VolumeGeometry),
            ("label", "i4"),
        ]
    )
    dtype = template_dtype

    def __getitem__(self, idx):
        """Implement dictionary access."""
        val = super().__getitem__(idx)
        if idx in ("src", "dst"):
            val = VolumeGeometry(val)
        return val

    def set_type(self, new_type):
        """
        Convert the internal transformation matrix to a different type inplace.

        Parameters
        ----------
        new_type : str, int
            Tranformation type

        """
        sa = self.structarr
        current = sa["type"]
        if isinstance(new_type, str):
            new_type = transform_codes.code[new_type]

        if current == new_type:
            return

        # VOX2VOX -> RAS2RAS
        if (current, new_type) == (0, 1):
            src = VolumeGeometry(sa["src"])
            dst = VolumeGeometry(sa["dst"])
            # See https://github.com/freesurfer/freesurfer/
            # blob/bbb2ef78591dec2c1ede3faea47f8dd8a530e92e/utils/transform.cpp#L3696-L3705
            # blob/bbb2ef78591dec2c1ede3faea47f8dd8a530e92e/utils/transform.cpp#L3548-L3568
            M = dst.as_affine() @ sa["m_L"] @ np.linalg.inv(src.as_affine())
            sa["m_L"] = M
            sa["type"] = new_type
            return

        raise NotImplementedError(
            "Converting {} to {} is not yet available".format(
                transform_codes.label[current], transform_codes.label[new_type]
            )
        )

    def to_ras(self, moving=None, reference=None):
        """
        Return a nitransforms' internal RAS+ array.

        Seemingly, the matrix of an LTA is defined such that it
        maps coordinates from the ``dest volume`` to the ``src volume``.
        Therefore, without inversion, the LTA matrix is appropiate
        to move the information from ``src volume`` into the
        ``dest volume``'s grid.

        .. important ::

            The ``moving`` and ``reference`` parameters are dismissed
            because ``VOX2VOX`` LTAs are converted to ``RAS2RAS`` type
            before returning the RAS+ matrix, using the ``dest`` and
            ``src`` contained in the LTA. Both arguments are kept for
            API compatibility.

        Parameters
        ----------
        moving : dismissed
            The spatial reference of moving images.
        reference : dismissed
            The spatial reference of moving images.

        Returns
        -------
        matrix : :obj:`numpy.ndarray`
            The RAS+ affine matrix corresponding to the LTA.

        """
        self.set_type(1)
        return np.linalg.inv(self.structarr["m_L"])

    def to_string(self, partial=False):
        """Convert this transform to text."""
        sa = self.structarr
        lines = (
            [
                "# LTA file created by NiTransforms",
                "type      = {}".format(sa["type"]),
                "nxforms   = 1",
            ]
            if not partial
            else []
        )

        # Standard preamble
        lines += [
            "mean      = {:6.4f} {:6.4f} {:6.4f}".format(*sa["mean"].flatten()),
            "sigma     = {:6.4f}".format(float(sa["sigma"])),
            "1 4 4",
        ]

        # Format parameters matrix
        lines += [" ".join(f"{v:18.15e}" for v in sa["m_L"][i]) for i in range(4)]

        lines += [
            "src volume info",
            str(self["src"]),
            "dst volume info",
            str(self["dst"]),
        ]

        lines += [] if partial else [""]
        return "\n".join(lines)

    @classmethod
    def from_string(cls, string, partial=False):
        """Read a transform from text."""
        lt = cls()
        sa = lt.structarr

        # Drop commented out lines
        lines = _drop_comments(string).splitlines()

        fields = ("type", "nxforms", "mean", "sigma")
        for key in fields[partial * 2:]:
            label, valstring = lines.pop(0).split(" = ")
            assert label.strip() == key

            if key != "nxforms":
                val = np.genfromtxt([valstring.encode()], dtype=cls.dtype[key])
                sa[key] = val.reshape(sa[key].shape)
            else:
                assert valstring.strip() == "1"

        assert lines.pop(0) == "1 4 4"  # xforms, shape + 1, shape + 1
        val = np.genfromtxt([valstring.encode() for valstring in lines[:4]], dtype="f4")
        sa["m_L"] = val
        lines = lines[4:]
        assert lines.pop(0) == "src volume info"
        sa["src"] = np.asanyarray(VolumeGeometry.from_string("\n".join(lines[:8])))
        lines = lines[8:]
        assert lines.pop(0) == "dst volume info"
        sa["dst"] = np.asanyarray(VolumeGeometry.from_string("\n".join(lines)))
        return lt

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an affine from a nitransform's RAS+ matrix."""
        lt = cls()
        sa = lt.structarr
        sa["sigma"] = 1.0
        sa["mean"] = np.zeros((3, 1), dtype="float")
        sa["type"] = 1  # RAS2RAS
        # Just for reference, nitransforms does not write VOX2VOX
        # PLEASE NOTE THAT LTA USES THE "POINTS" CONVENTION, therefore
        # the source is the reference (coordinates for which we need
        # to find a projection) and destination is the moving image
        # (from which data is pulled-back).
        if reference is not None:
            sa["src"] = np.asanyarray(VolumeGeometry.from_image(reference))

        if moving is not None:
            sa["dst"] = np.asanyarray(VolumeGeometry.from_image(moving))
        # However, the affine needs to be inverted
        # (i.e., it is not a pure "points" convention).
        # This inversion is consistent with self.to_ras()
        sa["m_L"] = np.linalg.inv(ras)
        # to make LTA file format
        return lt


class FSLinearTransformArray(BaseLinearTransformList):
    """A list of linear transforms generated by FreeSurfer."""

    template_dtype = np.dtype(
        [("type", "i4"), ("nxforms", "i4"), ("subject", "U1024"), ("fscale", "f4")]
    )
    dtype = template_dtype
    _inner_type = FSLinearTransform

    def __getitem__(self, idx):
        """Allow dictionary access to the transforms."""
        if idx == "xforms":
            return self._xforms
        if idx == "nxforms":
            return len(self._xforms)
        return self.structarr[idx]

    def to_ras(self, moving=None, reference=None):
        """Set type to RAS2RAS and return the new matrix."""
        self.structarr["type"] = 1
        return [xfm.to_ras(moving=moving, reference=reference) for xfm in self.xforms]

    def to_string(self):
        """Convert this LTA into text format."""
        code = int(self["type"])
        header = [
            "# LTA-array file created by NiTransforms",
            f"type      = {code} # {transform_codes.label[code]}",
            "nxforms   = {}".format(self["nxforms"]),
        ]
        xforms = [xfm.to_string(partial=True) for xfm in self._xforms]
        footer = [
            "subject {}".format(self["subject"]),
            "fscale {:.6f}".format(float(self["fscale"])),
            "",
        ]
        return "\n".join(header + xforms + footer)

    @classmethod
    def from_string(cls, string):
        """Read this LTA from a text string."""
        lta = cls()
        sa = lta.structarr

        # Drop commented out lines
        lines = _drop_comments(string).splitlines()
        if not lines or not lines[0].startswith("type"):
            raise TransformFileError("Invalid LTA format")

        for key in ("type", "nxforms"):
            label, valstring = lines.pop(0).split(" = ")
            assert label.strip() == key

            val = np.genfromtxt([valstring.encode()], dtype=cls.dtype[key])
            sa[key] = val.reshape(sa[key].shape) if val.size else ""
        for _ in range(sa["nxforms"]):
            lta._xforms.append(
                cls._inner_type.from_string("\n".join(lines[:25]), partial=True)
            )
            lta._xforms[-1].structarr["type"] = sa["type"]
            lines = lines[25:]
        for key in ("subject", "fscale"):
            # Optional keys
            if not (lines and lines[0].startswith(key)):
                continue
            try:
                label, valstring = lines.pop(0).split(" ")
            except ValueError:
                sa[key] = ""
            else:
                assert label.strip() == key

                val = np.genfromtxt([valstring.encode()], dtype=cls.dtype[key])
                sa[key] = val.reshape(sa[key].shape) if val.size else ""

        assert len(lta._xforms) == sa["nxforms"]
        return lta

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an affine from a nitransform's RAS+ matrix."""
        if ras.ndim == 2:
            return cls._inner_type.from_ras(ras, moving=moving, reference=reference)

        lt = cls()
        sa = lt.structarr
        sa["type"] = 1
        sa["nxforms"] = ras.shape[0]
        for i in range(sa["nxforms"]):
            lt._xforms.append(
                cls._inner_type.from_ras(
                    ras[i, ...], moving=moving, reference=reference
                )
            )

        sa["subject"] = "unset"
        sa["fscale"] = 0.0
        return lt


def _drop_comments(string):
    """Drop comments."""
    return "\n".join(
        [
            line.split("#")[0].strip()
            for line in string.splitlines()
            if line.split("#")[0].strip()
        ]
    )



# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def add_server_arguments(ap):
    """Defines common options for the Neuroglancer server."""
    g = ap.add_argument_group(title="Neuroglancer server options")
    g.add_argument(
        "--bind-address",
        help="Bind address for Python web server.  Use 127.0.0.1 (the default) to restrict access "
        "to browers running on the local machine, use 0.0.0.0 to permit access from remote browsers.",
    )
    g.add_argument(
        "--bind-port",
        help="Bind port for Python web server.  Use 0 (the default) to choose a port automatically.",
    )
    g.add_argument(
        "--static-content-url",
        help="Obtain the Neuroglancer client code from the specified URL.",
    )
    g.add_argument(
        "--static-content-dev-server",
        action="store_true",
        help="Build Neuroglancer client from source using dev server.",
    )
    g.add_argument(
        "--debug-server",
        action="store_true",
        help="Log requests to web server used for Neuroglancer Python API",
    )


def add_state_arguments(ap, required=False, dest="state"):
    """Defines options for specifying a Neuroglancer state."""
    g = ap.add_mutually_exclusive_group(required=required)

    def neuroglancer_url(s):
        from .url_state import parse_url

        return parse_url(s)

    g.add_argument(
        "--url",
        type=neuroglancer_url,
        dest=dest,
        help="Neuroglancer URL from which to obtain state.",
    )

    def json_state(path):
        import json

        from . import viewer_state

        with open(path) as f:
            return viewer_state.ViewerState(json.load(f))

    g.add_argument(
        "--json",
        type=json_state,
        dest=dest,
        help="Path to file containing Neuroglancer JSON state.",
    )


def handle_server_arguments(args):
    """Handles the options defined by `add_server_arguments`."""
    from . import server

    if args.bind_address is not None or args.bind_port is not None:
        server.set_server_bind_address(args.bind_address, args.bind_port)
    if args.static_content_dev_server:
        server.set_dev_server_content_source()
    elif args.static_content_url:
        server.set_static_content_source(url=args.static_content_url)
    if args.debug_server:
        server.debug = True



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common interface for transforms."""

import os
from collections.abc import Iterable
import numpy as np

from .base import (
    TransformBase,
    TransformError,
)
from .linear import (  # noqa: F401
    Affine,
    from_x5 as linear_from_x5,
)
from .nonlinear import (  # noqa: F401
    DenseFieldTransform,
    from_x5 as nonlinear_from_x5,
)

try:
    import h5py
    from .io import itk, x5 as x5io
    from .io.x5 import from_filename as load_x5
    _X5_AVAILABLE = True
except ImportError:
    _X5_AVAILABLE = False


class TransformChain(TransformBase):
    """Implements the concatenation of transforms."""

    __slots__ = ("_transforms",)

    def __init__(self, transforms=None):
        """Initialize a chain of transforms."""
        super().__init__()
        self._transforms = None

        if transforms is not None:
            self.transforms = transforms

    def __add__(self, b):
        """
        Compose this and other transforms.

        Example
        -------
        >>> T1 = TransformBase()
        >>> added = T1 + TransformBase() + TransformBase()
        >>> isinstance(added, TransformChain)
        True

        >>> len(added.transforms)
        3

        """
        self.append(b)
        return self

    def __getitem__(self, i):
        """
        Enable indexed access of transform chains.

        Example
        -------
        >>> T1 = TransformBase()
        >>> chain = T1 + TransformBase()
        >>> chain[0] is T1
        True

        """
        return self.transforms[i]

    def __len__(self):
        """Enable using len()."""
        return len(self.transforms)

    @property
    def ndim(self):
        """Get the number of dimensions."""
        return max(x.ndim for x in self._transforms)

    @property
    def transforms(self):
        """Get the internal list of transforms."""
        return self._transforms

    @transforms.setter
    def transforms(self, value):
        self._transforms = _as_chain(value)
        if self.transforms[0].reference:
            self.reference = self.transforms[0].reference

    def append(self, x):
        """
        Concatenate one element to the chain.

        Example
        -------
        >>> chain = TransformChain(transforms=TransformBase())
        >>> chain.append((TransformBase(), TransformBase()))
        >>> len(chain)
        3

        """
        self.transforms += _as_chain(x)

    def insert(self, i, x):
        """
        Insert an item at a given position.

        Example
        -------
        >>> chain = TransformChain(transforms=[TransformBase(), TransformBase()])
        >>> chain.insert(1, TransformBase())
        >>> len(chain)
        3

        >>> chain.insert(1, TransformChain(chain))
        >>> len(chain)
        6

        """
        self.transforms = self.transforms[:i] + _as_chain(x) + self.transforms[i:]

    def map(self, x, inverse=False):
        """
        Apply a succession of transforms, e.g., :math:`y = f_3(f_2(f_1(f_0(x))))`.

        Example
        -------
        >>> chain = TransformChain(transforms=[TransformBase(), TransformBase()])
        >>> chain([(0., 0., 0.), (1., 1., 1.), (-1., -1., -1.)])
        [(0.0, 0.0, 0.0), (1.0, 1.0, 1.0), (-1.0, -1.0, -1.0)]

        >>> chain([(0., 0., 0.), (1., 1., 1.), (-1., -1., -1.)], inverse=True)
        [(0.0, 0.0, 0.0), (1.0, 1.0, 1.0), (-1.0, -1.0, -1.0)]

        >>> TransformChain()((0., 0., 0.))  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        TransformError:

        """
        if not self.transforms:
            raise TransformError("Cannot apply an empty transforms chain.")

        transforms = self.transforms
        if inverse:
            transforms = list(reversed(self.transforms))

        for xfm in transforms:
            x = xfm.map(x, inverse=inverse)

        return x

    def asaffine(self, indices=None):
        """
        Combine a succession of linear transforms into one.

        Example
        ------
        >>> chain = TransformChain(transforms=[
        ...     Affine.from_matvec(vec=(2, -10, 3)),
        ...     Affine.from_matvec(vec=(-2, 10, -3)),
        ... ])
        >>> chain.asaffine()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        >>> chain = TransformChain(transforms=[
        ...     Affine.from_matvec(vec=(1, 2, 3)),
        ...     Affine.from_matvec(mat=[[0, 1, 0], [0, 0, 1], [1, 0, 0]]),
        ... ])
        >>> chain.asaffine()
        array([[0., 1., 0., 2.],
               [0., 0., 1., 3.],
               [1., 0., 0., 1.],
               [0., 0., 0., 1.]])

        >>> np.allclose(
        ...     chain.map((4, -2, 1)),
        ...     chain.asaffine().map((4, -2, 1)),
        ... )
        True

        Parameters
        ----------
        indices : :obj:`numpy.array_like`
            The indices of the values to extract.

        """
        affines = (
            self.transforms if indices is None else np.take(self.transforms, indices)
        )
        retval = affines[0]
        for xfm in affines[1:]:
            retval = xfm @ retval
        return retval

    @classmethod
    def from_filename(cls, filename, fmt="X5", reference=None, moving=None, x5_chain=0):
        """Load a transform file."""

        if not _X5_AVAILABLE:
            raise ImportWarning("h5py")

        retval = []
        if fmt and fmt.upper() == "X5":
            # Get list of X5 nodes and generate transforms
            xfm_list = [
                globals()[f"{node.type}_from_x5"]([node]) for node in load_x5(filename)
            ]
            if not xfm_list:
                raise TransformError("Empty transform group")

            if x5_chain is None:
                return xfm_list

            with h5py.File(str(filename), "r") as f:
                chain_grp = f.get("TransformChain")
                if chain_grp is None:
                    raise TransformError("X5 file contains no TransformChain")

                chain_path = chain_grp[str(x5_chain)][()]
                chain_path = (
                    chain_path.decode() if isinstance(chain_path, bytes) else chain_path
                )

            return TransformChain([xfm_list[int(idx)] for idx in chain_path.split("/")])

        if str(filename).endswith(".h5"):
            reference = None
            xforms = itk.ITKCompositeH5.from_filename(filename)
            for xfmobj in xforms:
                if isinstance(xfmobj, itk.ITKLinearTransform):
                    retval.insert(0, Affine(xfmobj.to_ras(), reference=reference))
                else:
                    retval.insert(0, DenseFieldTransform(xfmobj))

            return TransformChain(retval)

        raise NotImplementedError

    def to_filename(self, filename, fmt="X5"):
        """Store the transform chain in X5 format."""

        if not _X5_AVAILABLE:
            raise ImportWarning("h5py")

        if fmt.upper() != "X5":
            raise NotImplementedError("Only X5 format is supported for chains")

        existing = (
            self.from_filename(filename, x5_chain=None)
            if os.path.exists(filename)
            else []
        )

        xfm_chain = []
        new_xfms = []
        next_xfm_index = len(existing)
        for xfm in self.transforms:
            for eidx, existing_xfm in enumerate(existing):
                if xfm == existing_xfm:
                    xfm_chain.append(eidx)
                    break
            else:
                xfm_chain.append(next_xfm_index)
                new_xfms.append((next_xfm_index, xfm))
                existing.append(xfm)
                next_xfm_index += 1

        mode = "r+" if os.path.exists(filename) else "w"
        with h5py.File(str(filename), mode) as f:
            if "Format" not in f.attrs:
                f.attrs["Format"] = "X5"
                f.attrs["Version"] = np.uint16(1)

            tg = f.require_group("TransformGroup")
            for idx, node in new_xfms:
                g = tg.create_group(str(idx))
                x5io._write_x5_group(g, node.to_x5())

            cg = f.require_group("TransformChain")
            cg.create_dataset(str(len(cg)), data="/".join(str(i) for i in xfm_chain))

        return filename


def _as_chain(x):
    """Convert a value into a transform chain."""
    if isinstance(x, TransformChain):
        return x.transforms
    if isinstance(x, TransformBase):
        return [x]
    if isinstance(x, Iterable):
        return list(x)
    return [x]


load = TransformChain.from_filename



"""py.test configuration."""
import os
from pathlib import Path
import numpy as np
import nibabel as nb
import pytest
import tempfile

_testdir = Path(os.getenv("TEST_DATA_HOME", "~/.nitransforms/testdata")).expanduser()
_datadir = Path(__file__).parent / "tests" / "data"


@pytest.fixture(autouse=True)
def doctest_autoimport(doctest_namespace):
    """Make available some fundamental modules to doctest modules."""
    doctest_namespace["np"] = np
    doctest_namespace["nb"] = nb
    doctest_namespace["os"] = os
    doctest_namespace["Path"] = Path
    doctest_namespace["regress_dir"] = _datadir
    doctest_namespace["test_dir"] = _testdir

    tmpdir = tempfile.TemporaryDirectory()
    doctest_namespace["tmpdir"] = tmpdir.name

    testdata = np.zeros((11, 11, 11), dtype="uint8")
    nifti_fname = str(Path(tmpdir.name) / "test.nii.gz")
    nb.Nifti1Image(testdata, np.eye(4)).to_filename(nifti_fname)
    doctest_namespace["testfile"] = nifti_fname
    yield
    tmpdir.cleanup()


@pytest.fixture
def data_path():
    """Return the test data folder."""
    return _datadir


@pytest.fixture
def testdata_path():
    """Return the heavy test-data folder."""
    return _testdir


@pytest.fixture
def get_testdata():
    """Generate data in the requested orientation."""

    return _reorient(_testdir / "someones_anatomy.nii.gz")


@pytest.fixture
def get_testmask():
    """Generate data in the requested orientation."""
    return _reorient(_testdir / "someones_anatomy_brainmask.nii.gz")


def _reorient(path):
    """Reorient the input NIfTI file."""
    img = nb.load(path)
    imgaff = img.affine
    _data = {"RAS": img}
    newaff = imgaff.copy()
    newaff[0, 0] *= -1.0
    newaff[0, 3] = imgaff.dot(np.hstack((np.array(img.shape[:3]) - 1, 1.0)))[0]
    _data["LAS"] = nb.Nifti1Image(np.flip(np.asanyarray(img.dataobj), 0), newaff, img.header)
    _data["LAS"].set_data_dtype(img.get_data_dtype())
    newaff = imgaff.copy()
    newaff[0, 0] *= -1.0
    newaff[1, 1] *= -1.0
    newaff[:2, 3] = imgaff.dot(np.hstack((np.array(img.shape[:3]) - 1, 1.0)))[:2]
    _data["LPS"] = nb.Nifti1Image(
        np.flip(np.flip(np.asanyarray(img.dataobj), 0), 1), newaff, img.header
    )
    _data["LPS"].set_data_dtype(img.get_data_dtype())
    A = nb.volumeutils.shape_zoom_affine(
        img.shape, img.header.get_zooms(), x_flip=False
    )
    R = nb.affines.from_matvec(nb.eulerangles.euler2mat(x=0.09, y=0.001, z=0.001))
    newaff = R.dot(A)
    oblique_img = nb.Nifti1Image(np.asanyarray(img.dataobj), newaff, img.header)
    oblique_img.header.set_qform(newaff, 1)
    oblique_img.header.set_sform(newaff, 1)
    _data["oblique"] = oblique_img
    _data["oblique"].set_data_dtype(img.get_data_dtype())

    return _data



"""Type aliases."""
# ruff: noqa
# flake8: noqa

# stdlib
from collections.abc import Mapping, Sequence
from numbers import Number
from threading import RLock
from typing import Any, Generic, Literal, ParamSpec, TypeVar

# externals
import numpy as np

# optionals
try:
    import neuroglancer as ng
    import neuroglancer.json_wrappers as ngj
except ImportError:
    import ngtools._nglite as ng
    import ngtools._nglite.json_wrappers as ngj

T = TypeVar('T')
P = ParamSpec('P')

LayerDataSourceURL = (
    ng.LocalVolume |
    ng.skeleton.SkeletonSource |
    str
)
"""Type of `LayerDataSource.url`."""

LayerDataSourceLike = ng.LayerDataSource | LayerDataSourceURL | dict
"""Types that can be used to construct a `LayerDataSource`."""

LayerDataSourceTransform = ng.CoordinateSpaceTransform
"""Type of `LayerDataSource.transform`."""

LayerDataSubSources = dict[str, ng.LayerDataSubsource]
"""Type of `LayerDataSource.subsources`."""

SourceType = ng.LayerDataSource | ng.LocalVolume | ng.skeleton.SkeletonSource

LayerDataSourcesLike = (
    ng.LayerDataSources |
    LayerDataSourceLike |
    list[LayerDataSourceLike]
)
"""Types that can be used to construct a `LayerDataSourcesLike`."""

CoordinateSpaceTransformOutputDimensions = ng.CoordinateSpace
CoordinateSpaceTransformInputDimensions = ng.CoordinateSpace
CoordinateSpaceTransformSourceRank = int
CoordinateSpaceTransformMatrix = np.ndarray[np.float64]


# Less useful types

ToolType = str
ShaderControlToolControl = str
DimensionToolDimension = str
SidePanelLocationSide = str
SidePanelLocationVisible = bool
SidePanelLocationSize = int
SidePanelLocationFlex = float
SidePanelLocationRow = int
SidePanelLocationCol = int
SelectedLayerStateLayer = str
LayerSidePanelStateTab = str
LayerSidePanelStateTabs = set[str]
DimensionPlaybackVelocityVelocity = float
DimensionPlaybackVelocityAtBoundary = str
DimensionPlaybackVelocityPaused = bool
LayerType = str
LayerDimensions = ng.CoordinateSpace
LayerPosition = np.ndarray[np.float32]
LayerLocalVelocity = dict[str, ng.DimensionPlaybackVelocity]
LayerTab = str
LayerPanels = list[ng.LayerSidePanelState]
LayerPanel = ng.LayerSidePanelState
LayerPick = bool
LayerToolBindings = dict[str, ng.Tool]
LayerTool = ng.Tool
PointAnnotationLayerPoints = np.ndarray[np.float32]
ImageLayerSource = ng.LayerDataSources
ImageLayerShader = str
ShaderControlParameter = (
    ng.TransferFunctionParameters |
    ng.InvlerpParameters |
    Number |
    str
)
ShaderControls = dict[str, ShaderControlParameter]
ImageLayerShaderControls = ShaderControls
ImageLayerOpacity = float
ImageLayerBlend = str
ImageLayerVolumeRenderingMode = bool | str
ImageLayerVolumeRenderingGain = float
ImageLayerCrossSectionRenderScale = float
SegmentationLayerSource = ng.LayerDataSources
StarredSegmentsLike = ng.StarredSegments | dict[int, bool]
VisibleSegmentsLike = ng.VisibleSegments | set[int]
SegmentationLayerStarredSegments = ng.StarredSegments
SegmentationLayerVisibleSegments = ng.VisibleSegments
SegmentationLayerEquivalences = ng.EquivalenceMap


# ----------------------------------------------------------------------
# These types are not actually used.
# These are just my way to document the ng API.
# ----------------------------------------------------------------------

class Array(np.ndarray, Generic[T]):
    dtype: T


class SizedArray(tuple, Array):
    ...


class JsonObjectWrapper(ngj.JsonObjectWrapper):

    ArgType = ngj.JsonObjectWrapper | dict | None

    def __init__(self, arg: ArgType = None, **kwargs): ...

    _json_data: dict[str, object] = {}
    _cached_wrappers: dict[str, None | tuple[Any, Any]] = {}
    _lock: RLock = RLock()
    _readonly: bool = False

    # Class-specific properties:
    # property(
    #   fget: Gets attribute from cache and postprocess it using a type.
    #   fset: Pass value through validator and saves it in cache.
    # )

    def __setattr__(self, key: str, value: object) -> None:
        # When calling `setattr(key, value)`, the value to set is stored in
        # the cache, along with the corresponding value in `json_data``:
        # ```
        # _cached_wrappers[key] = (value, _json_data[key])
        # ```
        ...

    def __getattr__(self, key: str) -> object:
        # When calling `getattr(key)`, the cached value and json_data value
        # are retrieved.
        # * If the second value in the cache (i.e., the cached json_data
        #   value) and the value in json_data match (same address), it
        #   means that the json_data has not been updated since the set
        #   data was cached, and the latter can be returned.
        # * Otherwise, the data in json_data is returned, but first it
        #   is transformed into the correct type using the type wrapper
        #   and cached.
        # ```
        # current_json_data = _json_data[key]
        # (value, old_json_data) = _cached_wrappers[key]
        # if old_json_data is not current_json_data:
        #   value = wrapped_type(current_json_data)
        #   _cached_wrappers[key] = (value, current_json_data)
        # return value
        # ```
        ...


class Layers(ng.Layers):
    # Implements the `list[ng.ManagedLayer]` protocol.
    # Also implements the `dict[str, ng.ManagedLayer]` protocol.
    ArgType = (
        list[ng.ManagedLayer | dict] |
        dict[str, ng.Layer | ng.local_volume.LocalVolume | dict | None] |
        None
    )


class ManagedLayer(JsonObjectWrapper, ng.ManagedLayer):
    ArgNameType = str | ng.ManagedLayer
    ArgLayerType = ng.Layer | ng.local_volume.LocalVolume | dict | None
    # !! `layer` cannot be a `ManagedLayer``

    def __init__(
        self, name: ArgNameType, layer: ArgLayerType = None, **kwargs
    ) -> None: ...

    name: str
    layer: ng.Layer = ng.Layer()
    visible: bool | None = True
    archived: bool | None = False
    # All other attributes redirect to `layer`'s attributes.


class Layer(JsonObjectWrapper, ng.Layer):
    ArgType = ng.Layer | dict | None

    # kwargs / properties
    type: str | None = None
    layer_dimensions: ng.CoordinateSpace = ng.CoordinateSpace()
    layer_position: Array[np.float32] = np.ndarray([], np.float32)
    local_velocity: dict[str, ng.DimensionPlaybackVelocity]
    tab: str | None = None
    panels: list[ng.LayerSidePanelState] = []
    pick: bool = None
    tool_bindings: dict[str, ng.Tool] = {}
    tool: ng.Tool | None = None


ShaderControl = Number | ng.TransferFunctionParameters | ng.InvlerpParameters


class ImageLayer(Layer, ng.ImageLayer):
    ArgType = ng.ImageLayer | dict | None

    # kwargs / properties
    type: Literal["image"] = "image"
    source: ng.LayerDataSources = ng.LayerDataSources()
    shader: str = ""
    shader_controls: dict[str, ShaderControl] = {}
    opacity: float | None = 0.5
    blend: str | None = None
    volume_rendering_mode: bool | str | None = False
    volume_rendering_gain: float | None = 0.
    volume_rendering_depth_samples: float | None = 64.
    cross_section_render_scale: float | None = 1.
    annotation_color: str | None = None


class SegmentationLayer(Layer, ng.SegmentationLayer):
    ArgType = ng.SegmentationLayer | dict | None

    # kwargs / properties
    type: Literal["segmentation"] = "segmentation"
    source: ng.LayerDataSources = ng.LayerDataSources()
    starred_segments: ng.StarredSegments = ng.StarredSegments()
    visible_segments: ng.VisibleSegments = ng.VisibleSegments(starred_segments)
    segments: ng.VisibleSegments = visible_segments
    equivalences: ng.EquivalenceMap = ng.EquivalenceMap()
    hide_segment_zero: bool | None = True
    hover_highlight: bool | None = True
    base_segment_coloring: bool | None = False
    selected_alpha: float | None = 0.5
    not_selected_alpha: float | None = 0.
    object_alpha: float | None = 1.
    saturation: float | None = 1.
    ignore_null_visible_set: bool | None = True
    skeleton_rendering: ng.SkeletonRenderingOptions = ng.SkeletonRenderingOptions()
    skeleton_shader: str | None = skeleton_rendering.shader
    color_seed: int | None = 0
    cross_section_render_scale: float | None = 1.
    mesh_render_scale: float | None = 10.
    mesh_silhouette_rendering: float | None = 0.
    segment_query: str | None = None
    segment_colors: dict[np.uint64, str] = {}
    segment_default_color: str | None = None
    segment_html_color_dict: dict[np.uint64, str]
    linked_segmentation_group: str | None = None
    linked_segmentation_color_group: str | Literal[False] | None = None


class SkeletonRenderingOptions(JsonObjectWrapper, ng.SkeletonRenderingOptions):
    ArgType = ng.SkeletonRenderingOptions | dict | None

    # kwargs / properties
    shader: str | None = None
    shader_controls: dict[str, ShaderControl] = {}
    mode2d: str | None = None
    line_width2d: float | None = 2.
    mode3d: str | None = None
    line_width3d: float | None = 1.


class StarredSegments(JsonObjectWrapper, ng.StarredSegments, Mapping[np.uint64, bool]):
    # Implements the `dict[uint64, bool]` protocol

    ArgType = (
        None |
        ng.StarredSegments |
        Mapping[int | str, bool] |
        Sequence[tuple[int | str, bool]] |
        Sequence[int | str]
    )
    _data: dict[np.uint64, bool] = {}
    _visible: dict[np.uint64, bool] = {}


class AnnotationLayer(Layer, ng.AnnotationLayer):
    ArgType = ng.AnnotationLayer | dict | None

    type: Literal["annotation"] = "annotation"
    source: ng.LayerDataSources = ng.LayerDataSources()
    annotations: list[ng.AnnotationBase | dict] = []
    annotation_color: str | None = None
    annotation_properties: list[ng.AnnotationPropertySpec] = []
    annotation_relationships: list[str] = []
    linked_segmentation_layer: dict[str, str] = {}
    filter_by_segmentation: list[str] = []
    ignore_null_segment_filter: bool | None = True
    shader: str
    shader_controls: dict[str, ShaderControl]


class LocalAnnotationLayer(AnnotationLayer, ng.LocalAnnotationLayer):
    ArgType = ng.CoordinateSpace | dict  # dimensions
    Arg2Type = ng.LocalAnnotationLayer | dict | None

    source: ng.LayerDataSources = ng.LayerDataSources(url="local://annotations")


class AnnotationBase(JsonObjectWrapper, ng.AnnotationBase):
    ArgType = ng.viewer_state.AnnotationBase | dict | None

    id: str | None = None
    type: Literal["point", "line", "axis_aligned_bounding_box", "ellipsoid"]
    description: str | None = None
    segments: list[list[np.uint64]] | None = None
    props: list[Number | str] | None = None


class PointAnnotation(AnnotationBase, ng.PointAnnotation):
    ArgType = ng.PointAnnotation | dict | None
    type: Literal["point"] = "point"


class LineAnnotation(AnnotationBase, ng.LineAnnotation):
    ArgType = ng.LineAnnotation | dict | None
    type: Literal["line"] = "line"
    point_a: Array[np.float32]
    point_b: Array[np.float32]


class AxisAlignedBoundingBoxAnnotation(AnnotationBase, ng.AxisAlignedBoundingBoxAnnotation):
    ArgType = ng.AxisAlignedBoundingBoxAnnotation | dict | None
    type: Literal["axis_aligned_bounding_box"] = "axis_aligned_bounding_box"
    point_a: Array[np.float32]
    point_b: Array[np.float32]


class EllipsoidAnnotation(AnnotationBase, ng.EllipsoidAnnotation):
    ArgType = ng.EllipsoidAnnotation | dict | None
    type: Literal["ellipsoid"] = "ellipsoid"
    center: Array[np.float32]
    radii: Array[np.float32]


class AnnotationPropertySpec(JsonObjectWrapper, ng.AnnotationPropertySpec):
    ArgType = ng.AnnotationPropertySpec | dict | None
    id: str
    type: str
    description: str | None = None
    default: Number | str | None = None
    enum_values: list[Number | str] | None = None
    enum_values: list[str] | None = None


class InvlerpParameters(JsonObjectWrapper, ng.InvlerpParameters):
    ArgType = ng.InvlerpParameters | dict | None
    range: SizedArray[Number, Number] | None = None
    window: SizedArray[Number, Number] | None = None
    channel: list[int] | None = None


class TransferFunctionParameters(JsonObjectWrapper, ng.TransferFunctionParameters):
    ArgType = ng.TransferFunctionParameters | dict | None
    window: SizedArray[Number, Number] | None = None
    channel: list[int] | None = None
    controlPoints: list[list[Number | str]] | None = None
    defaultColor: str | None = None


class LayerDataSources(list[ng.LayerDataSource], ng.LayerDataSources):
    ArgType = (
        str |
        ng.local_volume.LocalVolume |
        ng.skeleton.SkeletonSource |
        ng.LayerDataSource |
        list[
            ng.local_volume.LocalVolume |
            ng.skeleton.SkeletonSource |
            ng.LayerDataSource |
            dict |
            None
        ] |
        ng.LayerDataSources |
        dict |
        None
    )

    # Implements the `list` protocol
    _data: list[ng.LayerDataSource] = []


class LayerDataSource(JsonObjectWrapper, ng.LayerDataSource):
    ArgType = (
        str |
        ng.local_volume.LocalVolume |
        ng.skeleton.SkeletonSource |
        ng.LayerDataSource |
        dict |
        None
    )

    url: (
        str |
        ng.local_volume.LocalVolume |
        ng.skeleton.SkeletonSource
    )
    transform: ng.CoordinateSpaceTransform | None = None
    subsources: dict[str, ng.LayerDataSubsource] = {}
    enable_default_subsources: bool | None = True


DataPanelLayoutType = Literal["xy", "yz", "xz", "xy-3d", "yz-3d", "xz-3d", "4panel", "3d"]
DataPanelLayoutLike = ng.DataPanelLayout | DataPanelLayoutType

StackLayoutType = Literal["row", "column"]
StackLayoutLike = ng.StackLayout | StackLayoutType

LayerGroupViewerType = Literal["viewer"]
LayerGroupViewerLike = LayerGroupViewerType | ng.LayerGroupViewer

LayoutSpecification = (
    DataPanelLayoutLike |
    StackLayoutLike |
    LayerGroupViewerLike |
    dict |
    None
)


class StackLayout(JsonObjectWrapper, ng.StackLayout):
    ArgType = ng.StackLayout | dict | None

    type: Literal["row", "column"]
    flex: float | None = 1.0
    children: list[LayoutSpecification]


class LayerGroupViewer(JsonObjectWrapper, ng.LayerGroupViewer):
    # Same view of multiple layers
    ArgType = ng.LayerGroupViewer | dict | None

    type: Literal["viewer"] = "viewer"
    flex: float | None = 1.0
    layers: list[str] = []
    layout: DataPanelLayoutLike = ng.DataPanelLayout("xy")
    position: Array[np.float32] = ng.LinkedPosition()
    velocity: dict[str, ng.DimensionPlaybackVelocity] = {}
    cross_section_orientation: Array[np.float32] = ng.LinkedOrientationState()
    cross_section_scale: float = ng.LinkedZoomFactor()
    cross_section_depth: float = ng.LinkedDepthRange()
    projection_orientation: Array[np.float32] = ng.LinkedOrientationState()
    projection_scale: float = ng.LinkedZoomFactor()
    projection_depth: float = ng.LinkedZoomFactor()
    tool_bindings: dict[str, ng.Tool] = {}


class DataPanelLayout(JsonObjectWrapper, ng.DataPanelLayout):
    ArgType = str | ng.DataPanelLayout | dict | None

    type: Literal["xy", "yz", "xz", "xy-3d", "yz-3d", "xz-3d", "4panel", "3d"]
    cross_sections : ng.CrossSectionMap
    orthographic_projection: bool | None = False


class CrossSectionMap(dict[str, ng.CrossSection], ng.CrossSectionMap):
    ArgType = dict[str, ng.CrossSection] | None


class CrossSection(JsonObjectWrapper, ng.CrossSection):
    ArgType = ng.CrossSection | dict | None

    width: int | None = 1000
    height: int | None = 1000
    position: Array[np.float32] = ng.LinkedPosition()
    orientation: Array[np.float32] = ng.LinkedOrientationState()
    scale: float = ng.LinkedZoomFactor()



"""Miscelaneous tools to build and manipulate Neuroglancer scenes."""
# ruff: noqa
# flake8: noqa
__all__ = ["__version__"]

# import to trigger fsspec registration
try:
    from . import dandifs as _
except ImportError:
    pass

# version
from ._version import __version__  # type: ignore

# monkey patch neuroglancer
from functools import wraps

try:
    from neuroglancer import StackLayout as _StackLayout
except ImportError:
    from ._nglite import StackLayout as _StackLayout

_old_stack_to_json = _StackLayout.to_json


@wraps(_old_stack_to_json)
def _stack_to_json(self: _StackLayout) -> dict:
    json = _old_stack_to_json(self)
    if "children" in json:
        json["children"] = [
            {"type": child} if isinstance(child, str) else child
            for child in json["children"]
        ]
    return json


_StackLayout.to_json = _stack_to_json



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

DEFAULT_MAX_DOWNSAMPLING = 64
DEFAULT_MAX_DOWNSAMPLED_SIZE = 128
DEFAULT_MAX_DOWNSAMPLING_SCALES = float("inf")


def compute_near_isotropic_downsampling_scales(
    size,
    voxel_size,
    dimensions_to_downsample,
    max_scales=DEFAULT_MAX_DOWNSAMPLING_SCALES,
    max_downsampling=DEFAULT_MAX_DOWNSAMPLING,
    max_downsampled_size=DEFAULT_MAX_DOWNSAMPLED_SIZE,
):
    """Compute a list of successive downsampling factors."""

    num_dims = len(voxel_size)
    cur_scale = np.ones((num_dims,), dtype=int)
    scales = [tuple(cur_scale)]
    while (
        len(scales) < max_scales
        and (np.prod(cur_scale) < max_downsampling)
        and (size / cur_scale).max() > max_downsampled_size
    ):
        # Find dimension with smallest voxelsize.
        cur_voxel_size = cur_scale * voxel_size
        smallest_cur_voxel_size_dim = dimensions_to_downsample[
            np.argmin(cur_voxel_size[dimensions_to_downsample])
        ]
        cur_scale[smallest_cur_voxel_size_dim] *= 2
        target_voxel_size = cur_voxel_size[smallest_cur_voxel_size_dim] * 2
        for d in dimensions_to_downsample:
            if d == smallest_cur_voxel_size_dim:
                continue
            d_voxel_size = cur_voxel_size[d]
            if abs(d_voxel_size - target_voxel_size) > abs(
                d_voxel_size * 2 - target_voxel_size
            ):
                cur_scale[d] *= 2
        scales.append(tuple(cur_scale))
    return scales


def compute_two_dimensional_near_isotropic_downsampling_scales(
    size,
    voxel_size,
    max_scales=float("inf"),
    max_downsampling=DEFAULT_MAX_DOWNSAMPLING,
    max_downsampled_size=DEFAULT_MAX_DOWNSAMPLED_SIZE,
):
    """Compute a list of successive downsampling factors for 2-d tiles."""

    max_scales = min(max_scales, 10)

    # First compute a set of 2-d downsamplings for XY, XZ, and YZ with a high
    # number of max_scales, and ignoring other criteria.
    scales_transpose = [
        compute_near_isotropic_downsampling_scales(
            size=size,
            voxel_size=voxel_size,
            dimensions_to_downsample=dimensions_to_downsample,
            max_scales=max_scales,
            max_downsampling=float("inf"),
            max_downsampled_size=0,
        )
        for dimensions_to_downsample in [[0, 1], [0, 2], [1, 2]]
    ]

    # Truncate all list of scales to the same length, once the stopping criteria
    # is reached for all values of dimensions_to_downsample.
    scales = [((1,) * 3,) * 3]
    size = np.array(size)

    def scale_satisfies_criteria(scale):
        return (
            np.prod(scale) < max_downsampling
            and (size / scale).max() > max_downsampled_size
        )

    for i in range(1, max_scales):
        cur_scales = tuple(scales_transpose[d][i] for d in range(3))
        if all(not scale_satisfies_criteria(scale) for scale in cur_scales):
            break
        scales.append(cur_scales)
    return scales
